# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from functools import cached_property
from uuid import UUID

from data_rentgen.db.utils.uuid import generate_static_uuid
from data_rentgen.dto.dataset import DatasetDTO
from data_rentgen.dto.dataset_column_relation import (
    DatasetColumnRelationDTO,
    merge_dataset_column_relations,
)
from data_rentgen.dto.operation import OperationDTO


@dataclass
class ColumnLineageDTO:
    operation: OperationDTO
    source_dataset: DatasetDTO
    target_dataset: DatasetDTO
    dataset_column_relations: InitVar[list[DatasetColumnRelationDTO]]
    _dataset_column_relations: list[DatasetColumnRelationDTO] = field(default_factory=list, init=False)
    # id is generated using other ids combination
    id: UUID | None = None

    def __post_init__(self, dataset_column_relations: list[DatasetColumnRelationDTO]):
        self._dataset_column_relations = merge_dataset_column_relations(dataset_column_relations)

    @property
    def unique_key(self) -> tuple:
        return (self.operation.unique_key, self.source_dataset.unique_key, self.target_dataset.unique_key)

    @property
    def column_relations(self) -> list[DatasetColumnRelationDTO]:
        return self._dataset_column_relations

    @cached_property
    def fingerprint(self) -> UUID:
        id_components = sorted((*item.unique_key, item.type) for item in self.column_relations)
        str_components = [".".join(map(str, item)) for item in id_components]
        return generate_static_uuid(",".join(str_components))

    def merge(self, new: ColumnLineageDTO) -> ColumnLineageDTO:
        return ColumnLineageDTO(
            operation=self.operation.merge(new.operation),
            source_dataset=self.source_dataset.merge(new.source_dataset),
            target_dataset=self.target_dataset.merge(new.target_dataset),
            dataset_column_relations=self.column_relations + new.column_relations,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Any

from data_rentgen.exceptions.base import ApplicationError


class RedirectError(ApplicationError):
    """Error which contains redirect url for authorization."""

    def __init__(self, message: str, details: Any = None) -> None:
        self._message = message
        self._details = details

    @property
    def message(self) -> str:
        return self._message

    @property
    def details(self) -> Any:
        return self._details



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from urllib.parse import urlparse

from data_rentgen.consumer.openlineage.job import OpenLineageJob
from data_rentgen.consumer.openlineage.run_facets import OpenLineageParentJob
from data_rentgen.dto import JobDTO, JobTypeDTO, LocationDTO


def extract_parent_job(job: OpenLineageParentJob) -> JobDTO:
    return JobDTO(
        name=job.name,
        location=extract_job_location(job),
    )


def extract_job(job: OpenLineageJob) -> JobDTO:
    return JobDTO(
        name=job.name,
        location=extract_job_location(job),
        type=extract_job_type(job),
    )


def extract_job_location(job: OpenLineageJob | OpenLineageParentJob) -> LocationDTO:
    url = urlparse(job.namespace)
    scheme = url.scheme or "unknown"
    netloc = url.netloc or url.path
    return LocationDTO(
        type=scheme,
        name=netloc,
        addresses={f"{scheme}://{netloc}"},
    )


def extract_job_type(job: OpenLineageJob) -> JobTypeDTO | None:
    if job.facets.jobType:
        job_type = job.facets.jobType.jobType
        integration_type = job.facets.jobType.integration
        return JobTypeDTO(f"{integration_type}_{job_type}")

    return None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from fastapi import Query
from pydantic import BaseModel, ConfigDict, Field

from data_rentgen.server.schemas.v1.location import LocationResponseV1
from data_rentgen.server.schemas.v1.pagination import PaginateQueryV1


class DatasetResponseV1(BaseModel):
    id: str = Field(description="Dataset id", coerce_numbers_to_str=True)
    location: LocationResponseV1 = Field(description="Corresponding Location")
    name: str = Field(description="Dataset name")
    format: str | None = Field(description="Data format", default=None)

    model_config = ConfigDict(from_attributes=True)


class DatasetDetailedResponseV1(BaseModel):
    id: str = Field(description="Dataset id", coerce_numbers_to_str=True)
    data: DatasetResponseV1 = Field(description="Dataset data")

    model_config = ConfigDict(from_attributes=True)


class DatasetPaginateQueryV1(PaginateQueryV1):
    """Query params for Dataset paginate request."""

    dataset_id: list[int] = Field(Query(default_factory=list), description="Dataset id")
    search_query: str | None = Field(
        Query(
            default=None,
            min_length=3,
            description="Search query",
        ),
    )

    model_config = ConfigDict(extra="forbid")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from faststream.security import BaseSecurity
from pydantic import BaseModel


class KafkaSecurityBaseSettings(BaseModel):
    def to_security(self) -> BaseSecurity:
        return BaseSecurity()

    def extra_broker_kwargs(self) -> dict[str, Any]:
        return {}

    async def initialize(self) -> None:
        return

    async def destroy(self) -> None:
        return

    async def refresh(self) -> None:
        return



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel, ConfigDict


class OpenLineageBase(BaseModel):
    """Base class for all OpenLineage models."""

    model_config = ConfigDict(extra="ignore", frozen=True, arbitrary_types_allowed=True)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from data_rentgen.server.middlewares.monitoring.metrics import (
    apply_monitoring_metrics_middleware,
)

__all__ = ["apply_monitoring_metrics_middleware"]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.job_facets.base import OpenLineageJobFacet
from data_rentgen.consumer.openlineage.job_facets.documentation import (
    OpenLineageDocumentationJobFacet,
)
from data_rentgen.consumer.openlineage.job_facets.job_type import (
    OpenLineageJobIntegrationType,
    OpenLineageJobProcessingType,
    OpenLineageJobType,
    OpenLineageJobTypeJobFacet,
)

__all__ = [
    "OpenLineageDocumentationJobFacet",
    "OpenLineageJobFacet",
    "OpenLineageJobFacets",
    "OpenLineageJobIntegrationType",
    "OpenLineageJobProcessingType",
    "OpenLineageJobType",
    "OpenLineageJobTypeJobFacet",
]


class OpenLineageJobFacets(OpenLineageBase):
    """All possible job facets.
    See [Job](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    documentation: OpenLineageDocumentationJobFacet | None = None
    jobType: OpenLineageJobTypeJobFacet | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import http
from typing import Any

from typing_extensions import Literal

from data_rentgen.server.errors.base import BaseErrorSchema
from data_rentgen.server.errors.registration import register_error_response


@register_error_response(
    exception=NotImplementedError,
    status=http.HTTPStatus.NOT_IMPLEMENTED,
)
class NotImplementedErrorSchema(BaseErrorSchema):
    code: Literal["not_implemented"] = "not_implemented"
    details: Any = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from uuid import UUID

from sqlalchemy import any_, select, text
from sqlalchemy.dialects.postgresql import insert

from data_rentgen.db.models import ColumnLineage
from data_rentgen.db.models.dataset_column_relation import (
    DatasetColumnRelation,
    DatasetColumnRelationType,
)
from data_rentgen.db.repositories.base import Repository
from data_rentgen.dto import ColumnLineageDTO


class DatasetColumnRelationRepository(Repository[ColumnLineage]):
    async def create_bulk_for_column_lineage(self, items: list[ColumnLineageDTO]) -> None:
        if not items:
            return

        # small optimization to avoid creating the same relations over and over.
        # although we're using ON CONFLICT DO NOTHING, PG still has to perform index scans
        # and to get next sequence value for each row
        fingerprints_to_create = await self._get_missing_fingerprints([item.fingerprint for item in items])
        relations_to_create = [item for item in items if item.fingerprint in fingerprints_to_create]

        if relations_to_create:
            await self._create_dataset_column_relations_bulk(relations_to_create)

    async def _get_missing_fingerprints(self, fingerprints: list[UUID]) -> set[UUID]:
        existing = await self._session.execute(
            select(DatasetColumnRelation.fingerprint.distinct()).where(
                DatasetColumnRelation.fingerprint == any_(fingerprints),  # type: ignore[arg-type]
            ),
        )

        return set(fingerprints) - set(existing.scalars().all())

    async def _create_dataset_column_relations_bulk(self, items: list[ColumnLineageDTO]):
        # we don't have to return anything, so there is no need to use o_conflict_update.
        # also rows are immutable, so there is no need to acquire any lock
        insert_statement = insert(DatasetColumnRelation).on_conflict_do_nothing(
            index_elements=[
                DatasetColumnRelation.fingerprint,
                DatasetColumnRelation.source_column,
                text("coalesce(target_column, '')"),
            ],
        )

        # several ColumnLineageDTO may have the same set of column relations, deduplicate them
        to_insert = defaultdict(list)
        for lineage in items:
            for column_relation in lineage.column_relations:
                key = (lineage.fingerprint, column_relation.source_column, column_relation.target_column)
                to_insert[key].append(column_relation)

        await self._session.execute(
            insert_statement,
            [
                {
                    # id is autoincremental
                    "fingerprint": fingerprint,
                    "source_column": relation.source_column,
                    "target_column": relation.target_column,
                    "type": DatasetColumnRelationType(relation.type).value,
                }
                for (fingerprint, *_), relations in to_insert.items()
                for relation in relations
            ],
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from typing import Any
from uuid import UUID

from data_rentgen.db.models.dataset_symlink import DatasetSymlink
from data_rentgen.db.models.operation import Operation
from data_rentgen.db.models.run import Run
from data_rentgen.db.repositories.column_lineage import ColumnLineageRow
from data_rentgen.db.repositories.input import InputRow
from data_rentgen.db.repositories.output import OutputRow
from data_rentgen.server.schemas.v1 import (
    ColumnLineageInteractionTypeV1,
    DatasetResponseV1,
    DirectLineageColumnRelationV1,
    IndirectLineageColumnRelationV1,
    JobResponseV1,
    LineageEntityKindV1,
    LineageEntityV1,
    LineageInputRelationV1,
    LineageNodesResponseV1,
    LineageOutputRelationV1,
    LineageParentRelationV1,
    LineageResponseV1,
    LineageSourceColumnV1,
    LineageSymlinkRelationV1,
    OperationResponseV1,
    RunResponseV1,
)
from data_rentgen.server.schemas.v1.lineage import (
    LineageIORelationSchemaV1,
    LineageRelationsResponseV1,
)
from data_rentgen.server.services.lineage import LineageServiceResult


async def build_lineage_response(lineage: LineageServiceResult) -> LineageResponseV1:
    datasets = {str(dataset.id): DatasetResponseV1.model_validate(dataset) for dataset in lineage.datasets.values()}
    jobs = {str(job.id): JobResponseV1.model_validate(job) for job in lineage.jobs.values()}
    runs = {run.id: RunResponseV1.model_validate(run) for run in lineage.runs.values()}
    operations = {op.id: OperationResponseV1.model_validate(op) for op in lineage.operations.values()}

    return LineageResponseV1(
        nodes=LineageNodesResponseV1(
            jobs=jobs,
            datasets=datasets,
            runs=runs,  # type: ignore[assignment, arg-type]
            operations=operations,  # type: ignore[assignment, arg-type]
        ),
        relations=LineageRelationsResponseV1(
            parents=_get_run_parent_relations(lineage.runs) + _get_operation_parent_relations(lineage.operations),
            symlinks=_get_symlink_relations(lineage.dataset_symlinks),
            inputs=_get_input_relations(lineage.inputs),
            outputs=_get_output_relations(lineage.outputs),
            direct_column_lineage=_get_direct_column_lineage(lineage.column_lineage),
            indirect_column_lineage=_get_indirect_column_lineage(lineage.column_lineage),
        ),
    )


def _get_run_parent_relations(runs: dict[UUID, Run]) -> list[LineageParentRelationV1]:
    parents = []
    for run_id in sorted(runs):
        run = runs[run_id]
        relation = LineageParentRelationV1(
            from_=LineageEntityV1(kind=LineageEntityKindV1.JOB, id=str(run.job_id)),
            to=LineageEntityV1(kind=LineageEntityKindV1.RUN, id=run.id),
        )
        parents.append(relation)
    return parents


def _get_operation_parent_relations(operations: dict[UUID, Operation]) -> list[LineageParentRelationV1]:
    parents = []
    for operation_id in sorted(operations):
        operation = operations[operation_id]
        relation = LineageParentRelationV1(
            from_=LineageEntityV1(kind=LineageEntityKindV1.RUN, id=operation.run_id),
            to=LineageEntityV1(kind=LineageEntityKindV1.OPERATION, id=operation.id),
        )
        parents.append(relation)
    return parents


def _get_symlink_relations(dataset_symlinks: dict[Any, DatasetSymlink]) -> list[LineageSymlinkRelationV1]:
    symlinks = []
    for key in sorted(dataset_symlinks):
        dataset_symlink = dataset_symlinks[key]
        relation = LineageSymlinkRelationV1(
            type=dataset_symlink.type,
            from_=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(dataset_symlink.from_dataset_id)),
            to=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(dataset_symlink.to_dataset_id)),
        )
        symlinks.append(relation)
    return symlinks


def _get_input_relations(inputs: dict[Any, InputRow]) -> list[LineageInputRelationV1]:
    relations = []
    for input_ in inputs.values():
        if input_.operation_id is not None:
            to = LineageEntityV1(kind=LineageEntityKindV1.OPERATION, id=input_.operation_id)
        elif input_.run_id is not None:
            to = LineageEntityV1(kind=LineageEntityKindV1.RUN, id=input_.run_id)
        elif input_.job_id is not None:
            to = LineageEntityV1(kind=LineageEntityKindV1.JOB, id=str(input_.job_id))

        relation = LineageInputRelationV1(
            from_=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(input_.dataset_id)),
            to=to,
            last_interaction_at=input_.created_at,
            num_bytes=input_.num_bytes,
            num_rows=input_.num_rows,
            num_files=input_.num_files,
            i_schema=LineageIORelationSchemaV1.model_validate(input_.schema) if input_.schema else None,
        )
        if relation.i_schema:
            relation.i_schema.relevance_type = input_.schema_relevance_type
        relations.append(relation)

    return sorted(relations, key=lambda x: (x.to.kind, str(x.from_.id), str(x.to.id)))


def _get_output_relations(outputs: dict[Any, OutputRow]) -> list[LineageOutputRelationV1]:
    relations = []
    for output in outputs.values():
        if output.operation_id is not None:
            from_ = LineageEntityV1(kind=LineageEntityKindV1.OPERATION, id=output.operation_id)
        elif output.run_id is not None:
            from_ = LineageEntityV1(kind=LineageEntityKindV1.RUN, id=output.run_id)
        elif output.job_id is not None:
            from_ = LineageEntityV1(kind=LineageEntityKindV1.JOB, id=str(output.job_id))

        relation = LineageOutputRelationV1(
            type=output.type,
            from_=from_,
            to=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(output.dataset_id)),
            last_interaction_at=output.created_at,
            num_bytes=output.num_bytes,
            num_rows=output.num_rows,
            num_files=output.num_files,
            o_schema=LineageIORelationSchemaV1.model_validate(output.schema) if output.schema else None,
        )
        if relation.o_schema:
            relation.o_schema.relevance_type = output.schema_relevance_type
        relations.append(relation)

    return sorted(relations, key=lambda x: (x.from_.kind, str(x.from_.id), str(x.to.id), x.type))


def _get_direct_column_lineage(column_lineage_by_source_target_id: dict[tuple, list[ColumnLineageRow]]):
    relations = []
    for (source_dataset_id, target_dataset_id), column_relations in column_lineage_by_source_target_id.items():
        column_lineage_relation = DirectLineageColumnRelationV1(
            from_=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(source_dataset_id)),
            to=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(target_dataset_id)),
        )
        fields = defaultdict(list)
        for column_relation in column_relations:
            if column_relation.target_column:
                fields[column_relation.target_column].append(
                    LineageSourceColumnV1(
                        field=column_relation.source_column,
                        last_used_at=column_relation.last_used_at,
                        types=[
                            type_
                            for type_ in ColumnLineageInteractionTypeV1
                            if type_.value & column_relation.types_combined
                        ],
                    ),
                )
        if fields:
            column_lineage_relation.fields = fields  # type: ignore[assignment]
            relations.append(column_lineage_relation)
    return sorted(relations, key=lambda x: (str(x.from_.id), str(x.to.id)))


def _get_indirect_column_lineage(column_lineage_by_source_target_id: dict[tuple, list[ColumnLineageRow]]):
    relations = []
    for (source_dataset_id, target_dataset_id), column_relations in column_lineage_by_source_target_id.items():
        column_lineage_relation = IndirectLineageColumnRelationV1(
            from_=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(source_dataset_id)),
            to=LineageEntityV1(kind=LineageEntityKindV1.DATASET, id=str(target_dataset_id)),
        )
        fields = [
            LineageSourceColumnV1(
                field=column_relation.source_column,
                last_used_at=column_relation.last_used_at,
                types=[
                    type_ for type_ in ColumnLineageInteractionTypeV1 if type_.value & column_relation.types_combined
                ],
            )
            for column_relation in column_relations
            if column_relation.target_column is None
        ]
        if fields:
            column_lineage_relation.fields = fields
            relations.append(column_lineage_relation)
    return sorted(relations, key=lambda x: (str(x.from_.id), str(x.to.id)))



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated, Any
from uuid import UUID as PlainUUID  # noqa: N811

from pydantic import PlainValidator
from uuid6 import UUID as UUIDv7  # noqa: N811


def uuid_version_validator(run_id: Any) -> UUIDv7:
    if isinstance(run_id, str):
        run_id = UUIDv7(run_id)
        if not run_id.version or run_id.version < 6:  # noqa: PLR2004
            err_msg = f"Run ID: {run_id} is not valid uuid. Only UUIDv6+ are supported"
            raise ValueError(err_msg)
        return run_id
    return run_id


# Teach Pydantic how to parse and represent UUID v7
# Right now use uuid from uuid lib cause: https://github.com/tiangolo/fastapi/issues/10259
UUIDv6Plus = Annotated[
    PlainUUID,
    PlainValidator(uuid_version_validator),
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from data_rentgen.dto.dataset import DatasetDTO


class DatasetSymlinkTypeDTO(str, Enum):
    METASTORE = "METASTORE"
    WAREHOUSE = "WAREHOUSE"

    def __str__(self) -> str:
        return self.value


@dataclass
class DatasetSymlinkDTO:
    from_dataset: DatasetDTO
    to_dataset: DatasetDTO
    type: DatasetSymlinkTypeDTO
    id: int | None = field(default=None, compare=False)

    @property
    def unique_key(self) -> tuple:
        return (self.from_dataset.unique_key, self.to_dataset.unique_key, self.type)

    def merge(self, new: DatasetSymlinkDTO) -> DatasetSymlinkDTO:
        return DatasetSymlinkDTO(
            from_dataset=self.from_dataset.merge(new.from_dataset),
            to_dataset=self.to_dataset.merge(new.to_dataset),
            type=self.type,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from urllib.parse import urlparse

from data_rentgen.consumer.openlineage.dataset import OpenLineageDataset
from data_rentgen.consumer.openlineage.dataset_facets import (
    OpenLineageStorageDatasetFacet,
    OpenLineageSymlinkIdentifier,
    OpenLineageSymlinkType,
)
from data_rentgen.consumer.openlineage.dataset_facets.column_lineage import (
    OpenLineageColumnLineageDatasetFacetFieldRef,
)
from data_rentgen.dto import (
    DatasetDTO,
    DatasetSymlinkDTO,
    DatasetSymlinkTypeDTO,
    LocationDTO,
)

logger = logging.getLogger(__name__)

OpenLineageDatasetLike = (
    OpenLineageDataset | OpenLineageSymlinkIdentifier | OpenLineageColumnLineageDatasetFacetFieldRef
)
METASTORE = DatasetSymlinkTypeDTO.METASTORE
WAREHOUSE = DatasetSymlinkTypeDTO.WAREHOUSE

PARTITION_PATH_PATTERN: re.Pattern = re.compile("^(.*?)/[^/=]+=")


def connect_dataset_with_symlinks(
    dataset: DatasetDTO,
    symlink: DatasetDTO,
    type_: OpenLineageSymlinkType,
) -> list[DatasetSymlinkDTO]:
    result = []
    is_metastore_symlink = type_ == OpenLineageSymlinkType.TABLE

    result.append(
        DatasetSymlinkDTO(
            from_dataset=dataset,
            to_dataset=symlink,
            type=METASTORE if is_metastore_symlink else WAREHOUSE,
        ),
    )
    result.append(
        DatasetSymlinkDTO(
            from_dataset=symlink,
            to_dataset=dataset,
            type=WAREHOUSE if is_metastore_symlink else METASTORE,
        ),
    )

    return sorted(result, key=lambda x: x.type)


def strip_partitions_from_path(name: str):
    # convert /some/long/path/with=partition/another=abc to /some/long/path
    if "=" not in name or "/" not in name:
        return name

    name_with_partitions = PARTITION_PATH_PATTERN.match(name)
    return name_with_partitions.group(1) if name_with_partitions else name


def extract_dataset_ref(dataset: OpenLineageDatasetLike) -> DatasetDTO:
    return DatasetDTO(
        name=strip_partitions_from_path(dataset.name),
        location=extract_dataset_location(dataset),
    )


def extract_dataset(dataset: OpenLineageDataset) -> DatasetDTO:
    dataset_dto = extract_dataset_ref(dataset)
    dataset_dto.format = extract_dataset_format(dataset)
    return dataset_dto


def extract_dataset_and_symlinks(dataset: OpenLineageDataset) -> tuple[DatasetDTO, list[DatasetSymlinkDTO]]:
    dataset_dto = extract_dataset(dataset)
    if not dataset.facets.symlinks:
        return dataset_dto, []

    table_symlinks = [
        identifier
        for identifier in dataset.facets.symlinks.identifiers
        if identifier.type == OpenLineageSymlinkType.TABLE
    ]
    if table_symlinks:
        # We are swapping the dataset with its TABLE symlink to create a cleaner lineage.
        # For example, by replacing an HDFS file with its corresponding Hive table.
        # This ensures that all operations interact with a single table instead of multiple files (which may represent different partitions).  # noqa: E501
        # Discussion on this issue: https://github.com/OpenLineage/OpenLineage/issues/2718

        # TODO: add support for multiple TABLE symlinks
        if len(table_symlinks) > 1:
            logger.warning(
                "Dataset has more than one TABLE symlink. "
                "Only the first one will be used for replacement. Symlink name: %s",
                table_symlinks[0].name,
            )
        table_dataset_dto = extract_dataset_ref(table_symlinks[0])
        return (
            table_dataset_dto,
            connect_dataset_with_symlinks(
                dataset_dto,
                table_dataset_dto,
                OpenLineageSymlinkType.TABLE,
            ),
        )

    symlinks = []
    for symlink_identifier in dataset.facets.symlinks.identifiers:
        symlink_dto = extract_dataset_ref(symlink_identifier)
        symlinks.extend(
            connect_dataset_with_symlinks(
                dataset_dto,
                symlink_dto,
                symlink_identifier.type,
            ),
        )

    return dataset_dto, symlinks


def extract_dataset_location(dataset: OpenLineageDatasetLike) -> LocationDTO:
    namespace = dataset.namespace
    if namespace == "file":
        # TODO: remove after https://github.com/OpenLineage/OpenLineage/issues/2709
        namespace = "file://"

    url = urlparse(namespace)
    scheme = url.scheme or "unknown"

    # TODO: handle S3 bucket properly after https://github.com/OpenLineage/OpenLineage/issues/2816
    netloc = url.netloc or url.path
    hosts = list(filter(None, netloc.split(","))) or ["unknown"]
    return LocationDTO(
        type=scheme,
        name=hosts[0],
        addresses={f"{scheme}://{host}" for host in hosts},
    )


def extract_dataset_format(dataset: OpenLineageDataset) -> str | None:
    match dataset.facets.storage:
        case OpenLineageStorageDatasetFacet(storageLayer="default", fileFormat=file_format):
            # See https://github.com/OpenLineage/OpenLineage/issues/2770
            return file_format
        case OpenLineageStorageDatasetFacet(storageLayer=storage_layer):
            return storage_layer
        case _:
            return None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from sqlalchemy import BigInteger, Computed, ForeignKey, Index, String, UniqueConstraint
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import Mapped, mapped_column, relationship

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.location import Location


class Dataset(Base):
    __tablename__ = "dataset"
    __table_args__ = (
        UniqueConstraint("location_id", "name"),
        Index("ix__dataset__search_vector", "search_vector", postgresql_using="gin"),
    )

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)

    location_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("location.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
        doc="Where dataset's data is actually located (database address, filesystem address)",
    )
    location: Mapped[Location] = relationship(
        Location,
        lazy="noload",
        foreign_keys=[location_id],
    )

    name: Mapped[str] = mapped_column(
        String,
        index=True,
        nullable=False,
        doc="Dataset name, e.g. table name or filesystem path",
    )
    format: Mapped[str | None] = mapped_column(
        String(32),
        nullable=True,
        doc="Data format, if any",
    )

    search_vector: Mapped[str] = mapped_column(
        TSVECTOR,
        Computed(
            # Postgres treated values like `mydb.mytable` as a whole word (domain name),
            # which does not allow user to search by name parts like `mytable`.
            # Same for slashes which are treated like file paths.
            # Keep both original name and one without punctuation to allow both full match and partial match.
            #
            # Also 'english' dictionary performs stemming,
            # so name like 'my.database.table' is converted to tsvector `'databas':2 'tabl':3`,
            # which does not match a tsquery like 'my:* & database:* & table:*'.
            # Instead prefer 'simple' dictionary as it does not use stemming.
            "to_tsvector('simple'::regconfig, name || ' ' || (translate(name, '/.', '  ')))",
            persisted=True,
        ),
        nullable=False,
        deferred=True,
        doc="Full-text search vector",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel, ConfigDict, Field


class UserResponseV1(BaseModel):
    name: str = Field(description="User name")
    model_config = ConfigDict(from_attributes=True)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from enum import Enum

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageSymlinkType(str, Enum):
    """Dataset symlink type.
    See [SymlinksDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/SymlinksDatasetFacet.json).
    """

    TABLE = "TABLE"
    LOCATION = "LOCATION"


class OpenLineageSymlinkIdentifier(OpenLineageBase):
    """Dataset symlink information.
    See [SymlinksDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/SymlinksDatasetFacet.json).
    """

    namespace: str
    name: str
    type: OpenLineageSymlinkType


class OpenLineageSymlinksDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing symlinks.
    See [SymlinksDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/SymlinksDatasetFacet.json).
    """

    identifiers: list[OpenLineageSymlinkIdentifier]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from data_rentgen.server.providers.auth.base_provider import AuthProvider
from data_rentgen.server.providers.auth.dummy_provider import DummyAuthProvider
from data_rentgen.server.providers.auth.keycloak_provider import KeycloakAuthProvider

__all__ = [
    "AuthProvider",
    "DummyAuthProvider",
    "KeycloakAuthProvider",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import dataclasses
import logging
from collections import defaultdict
from collections.abc import Collection
from dataclasses import dataclass, field
from datetime import datetime
from typing import Annotated, Literal
from uuid import UUID

from fastapi import Depends

from data_rentgen.db.models import (
    Dataset,
    DatasetSymlink,
    Job,
    Operation,
    Run,
)
from data_rentgen.db.repositories.column_lineage import ColumnLineageRow
from data_rentgen.db.repositories.input import InputRow
from data_rentgen.db.repositories.output import OutputRow
from data_rentgen.server.schemas.v1.lineage import LineageDirectionV1
from data_rentgen.services.uow import UnitOfWork

logger = logging.getLogger(__name__)


@dataclass
class LineageServiceResult:
    jobs: dict[int, Job] = field(default_factory=dict)
    runs: dict[UUID, Run] = field(default_factory=dict)
    operations: dict[UUID, Operation] = field(default_factory=dict)
    datasets: dict[int, Dataset] = field(default_factory=dict)
    dataset_symlinks: dict[tuple[int, int], DatasetSymlink] = field(default_factory=dict)
    inputs: dict[tuple[int, int, UUID | None, UUID | None], InputRow] = field(default_factory=dict)
    outputs: dict[tuple[int, int, UUID | None, UUID | None, str | None], OutputRow] = field(default_factory=dict)
    column_lineage: dict[tuple[int, int], list[ColumnLineageRow]] = field(default_factory=dict)

    def merge(self, other: "LineageServiceResult") -> "LineageServiceResult":
        self.jobs.update(other.jobs)
        self.runs.update(other.runs)
        self.operations.update(other.operations)
        self.datasets.update(other.datasets)
        self.dataset_symlinks.update(other.dataset_symlinks)
        self.inputs.update(other.inputs)
        self.outputs.update(other.outputs)
        self.column_lineage.update(other.column_lineage)
        return self


@dataclass
class IdsToSkip:
    jobs: set[int] = field(default_factory=set)
    runs: set[UUID] = field(default_factory=set)
    operations: set[UUID] = field(default_factory=set)
    datasets: set[int] = field(default_factory=set)

    @classmethod
    def from_result(cls, result: LineageServiceResult) -> "IdsToSkip":
        return cls(
            jobs=set(result.jobs.keys()),
            runs=set(result.runs.keys()),
            operations=set(result.operations.keys()),
            datasets=set(result.datasets.keys()),
        )

    def merge(self, other: "IdsToSkip") -> "IdsToSkip":
        self.jobs.update(other.jobs)
        self.runs.update(other.runs)
        self.operations.update(other.operations)
        self.datasets.update(other.datasets)
        return self


class LineageService:
    def __init__(self, uow: Annotated[UnitOfWork, Depends()]):
        self._uow = uow

    async def get_lineage_by_jobs(  # noqa: C901, PLR0912, PLR0915
        self,
        start_node_ids: Collection[int],
        direction: LineageDirectionV1,
        granularity: Literal["JOB", "RUN", "OPERATION"],
        since: datetime,
        until: datetime | None,
        depth: int,
        include_column_lineage: bool = False,  # noqa: FBT001, FBT002
        ids_to_skip: IdsToSkip | None = None,
        level: int = 0,
    ) -> LineageServiceResult:
        if not start_node_ids:
            return LineageServiceResult()

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Get lineage by runs %r, with direction %s, since %s, until %s",
                level,
                sorted(start_node_ids),
                direction,
                since,
                until,
            )

        jobs = await self._uow.job.list_by_ids(start_node_ids)
        if not jobs:
            logger.info("[Level %d] No jobs found", level)
            return LineageServiceResult()

        # Always include all requested jobs.
        jobs_by_id = {job.id: job for job in jobs}

        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            outputs.extend(
                await self._uow.output.list_by_job_ids(
                    jobs_by_id,
                    since=since,
                    until=until,
                    granularity=granularity,
                ),
            )
            if granularity != "JOB":
                # include sum over all runs
                outputs.extend(
                    await self._uow.output.list_by_job_ids(
                        jobs_by_id,
                        since=since,
                        until=until,
                        granularity="JOB",
                    ),
                )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            inputs.extend(
                await self._uow.input.list_by_job_ids(
                    jobs_by_id,
                    since=since,
                    until=until,
                    granularity=granularity,
                ),
            )
            if granularity != "JOB":
                # include sum over all runs
                inputs.extend(
                    await self._uow.input.list_by_job_ids(
                        jobs_by_id,
                        since=since,
                        until=until,
                        granularity="JOB",
                    ),
                )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}

        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        # Include only runs which have at least one input or output.
        # In case granularity == "JOB", all run_id are None.
        ids_to_skip = ids_to_skip or IdsToSkip()
        input_run_ids = {input_.run_id for input_ in inputs if input_.run_id is not None}
        output_run_ids = {output.run_id for output in outputs if output.run_id is not None}
        run_ids = input_run_ids | output_run_ids - ids_to_skip.runs
        runs = await self._uow.run.list_by_ids(run_ids)
        runs_by_id = {run.id: run for run in runs}

        result = LineageServiceResult(
            jobs=jobs_by_id,
            runs=runs_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        upstream_dataset_ids = {input_.dataset_id for input_ in inputs} - ids_to_skip.datasets
        downstream_dataset_ids = {output.dataset_id for output in outputs} - ids_to_skip.datasets
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `dataset1 -> current_job -> dataset2``,
            # but do not include relations like `another_job -> dataset2`.
            # Also datasets and symlinks will be populated by nested calls.
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=upstream_dataset_ids,
                    direction=LineageDirectionV1.UPSTREAM,
                    granularity=granularity,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=downstream_dataset_ids,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    granularity=granularity,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            # datasets and symlinks will be populated by nested call
            dataset_ids = upstream_dataset_ids | downstream_dataset_ids
            datasets = await self._uow.dataset.list_by_ids(dataset_ids)

            extra_datasets, dataset_symlinks = await self._extend_datasets_with_symlinks(dataset_ids, ids_to_skip)
            datasets.extend(extra_datasets)

            result.datasets = {dataset.id: dataset for dataset in datasets}
            result.dataset_symlinks = {
                (dataset_symlink.from_dataset_id, dataset_symlink.to_dataset_id): dataset_symlink
                for dataset_symlink in dataset_symlinks
            }

        if level == 0:
            if include_column_lineage:
                result.column_lineage.update(
                    await self._get_column_lineage(
                        current_result=result,
                        since=since,
                        until=until,
                        granularity="JOB",
                    ),
                )

            result = self._drop_unused_datasets(result)

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Found %d jobs, %d runs, %d operations, %d datasets, %d dataset symlinks, "
                "%d inputs, %d outputs, %d column lineage",
                level,
                len(result.jobs),
                len(result.runs),
                len(result.operations),
                len(result.datasets),
                len(result.dataset_symlinks),
                len(result.inputs),
                len(result.outputs),
                len(result.column_lineage),
            )
        return result

    async def get_lineage_by_runs(  # noqa: C901, PLR0915, PLR0912
        self,
        start_node_ids: Collection[UUID],
        direction: LineageDirectionV1,
        granularity: Literal["OPERATION", "RUN"],
        since: datetime,
        until: datetime | None,
        depth: int,
        include_column_lineage: bool = False,  # noqa: FBT001, FBT002
        ids_to_skip: IdsToSkip | None = None,
        level: int = 0,
    ) -> LineageServiceResult:
        if not start_node_ids:
            return LineageServiceResult()

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Get lineage by runs %r, with direction %s, since %s, until %s",
                level,
                sorted(start_node_ids),
                direction,
                since,
                until,
            )

        runs = await self._uow.run.list_by_ids(start_node_ids)
        if not runs:
            logger.info("[Level %d] No runs found", level)
            return LineageServiceResult()

        # Always include all requested runs.
        runs_by_id = {run.id: run for run in runs}

        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            outputs.extend(
                await self._uow.output.list_by_run_ids(
                    runs_by_id,
                    since=since,
                    until=until,
                    granularity=granularity,
                ),
            )
            if granularity != "RUN":
                # include sum over all operation
                outputs.extend(
                    await self._uow.output.list_by_run_ids(
                        runs_by_id,
                        since=since,
                        until=until,
                        granularity="RUN",
                    ),
                )
            # include sum over all runs
            outputs.extend(
                await self._uow.output.list_by_run_ids(
                    runs_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            inputs.extend(
                await self._uow.input.list_by_run_ids(
                    runs_by_id,
                    since=since,
                    until=until,
                    granularity=granularity,
                ),
            )
            if granularity != "RUN":
                # include sum over all operation
                inputs.extend(
                    await self._uow.input.list_by_run_ids(
                        runs_by_id,
                        since=since,
                        until=until,
                        granularity="RUN",
                    ),
                )
            # include sum over all runs
            inputs.extend(
                await self._uow.input.list_by_run_ids(
                    runs_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}

        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        # Include only operations which have at least one input or output.
        # In case granularity == "RUN", all operation_id are None.
        ids_to_skip = ids_to_skip or IdsToSkip()
        input_operation_ids = {input_.operation_id for input_ in inputs if input_.operation_id is not None}
        output_operation_ids = {output.operation_id for output in outputs if output.operation_id is not None}
        operation_ids = input_operation_ids | output_operation_ids - ids_to_skip.operations
        operations = await self._uow.operation.list_by_ids(operation_ids)
        operations_by_id = {operation.id: operation for operation in operations}

        job_ids = {run.job_id for run in runs} - ids_to_skip.jobs
        jobs = await self._uow.job.list_by_ids(job_ids)
        jobs_by_id = {job.id: job for job in jobs}

        result = LineageServiceResult(
            jobs=jobs_by_id,
            runs=runs_by_id,
            operations=operations_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        upstream_dataset_ids = {input_.dataset_id for input_ in inputs} - ids_to_skip.datasets
        downstream_dataset_ids = {output.dataset_id for output in outputs} - ids_to_skip.datasets
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `dataset1 -> current_run -> dataset2``,
            # but do not include relations like `another_run -> dataset2`.
            # Also datasets and symlinks will be populated by nested calls.
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=upstream_dataset_ids,
                    direction=LineageDirectionV1.UPSTREAM,
                    granularity=granularity,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=downstream_dataset_ids,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    granularity=granularity,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            dataset_ids = upstream_dataset_ids | downstream_dataset_ids
            datasets = await self._uow.dataset.list_by_ids(dataset_ids)

            extra_datasets, dataset_symlinks = await self._extend_datasets_with_symlinks(dataset_ids, ids_to_skip)
            datasets.extend(extra_datasets)

            result.datasets = {dataset.id: dataset for dataset in datasets}
            result.dataset_symlinks = {
                (dataset_symlink.from_dataset_id, dataset_symlink.to_dataset_id): dataset_symlink
                for dataset_symlink in dataset_symlinks
            }

        if level == 0:
            if include_column_lineage:
                result.column_lineage.update(
                    await self._get_column_lineage(current_result=result, since=since, until=until, granularity="RUN"),
                )

            result = self._drop_unused_datasets(result)

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Found %d jobs, %d runs, %d operations, %d datasets, %d dataset symlinks, "
                "%d inputs, %d outputs, %d column lineage",
                level,
                len(result.jobs),
                len(result.runs),
                len(result.operations),
                len(result.datasets),
                len(result.dataset_symlinks),
                len(result.inputs),
                len(result.outputs),
                len(result.column_lineage),
            )
        return result

    async def get_lineage_by_operations(  # noqa: C901, PLR0912, PLR0915
        self,
        start_node_ids: Collection[UUID],
        direction: LineageDirectionV1,
        since: datetime,
        until: datetime | None,
        depth: int,
        include_column_lineage: bool = False,  # noqa: FBT001, FBT002
        ids_to_skip: IdsToSkip | None = None,
        level: int = 0,
    ) -> LineageServiceResult:
        if not start_node_ids:
            return LineageServiceResult()

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Get lineage by operations %r, with direction %s, since %s, until %s",
                level,
                sorted(start_node_ids),
                direction,
                since,
                until,
            )

        operations = await self._uow.operation.list_by_ids(start_node_ids)
        if not operations:
            logger.info("[Level %d] No operations found", level)
            return LineageServiceResult()

        # Always include all requested operations.
        operations_by_id = {operation.id: operation for operation in operations}

        # Also include all parent runs & jobs.
        ids_to_skip = ids_to_skip or IdsToSkip()
        run_ids = {operation.run_id for operation in operations}
        runs = await self._uow.run.list_by_ids(run_ids - ids_to_skip.runs)
        runs_by_id = {run.id: run for run in runs}

        job_ids = {run.job_id for run in runs}
        jobs = await self._uow.job.list_by_ids(job_ids - ids_to_skip.jobs)
        jobs_by_id = {job.id: job for job in jobs}

        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            outputs.extend(
                await self._uow.output.list_by_operation_ids(operations_by_id, granularity="OPERATION"),
            )
            # include sum over all operations
            outputs.extend(
                await self._uow.output.list_by_operation_ids(operations_by_id, granularity="RUN"),
            )
            # include sum over all runs
            outputs.extend(
                await self._uow.output.list_by_operation_ids(operations_by_id, granularity="JOB"),
            )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            inputs.extend(
                await self._uow.input.list_by_operation_ids(operations_by_id, granularity="OPERATION"),
            )
            # include sum over all operations
            inputs.extend(
                await self._uow.input.list_by_operation_ids(operations_by_id, granularity="RUN"),
            )
            # include sum over all runs
            inputs.extend(
                await self._uow.input.list_by_operation_ids(operations_by_id, granularity="JOB"),
            )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}

        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        result = LineageServiceResult(
            jobs=jobs_by_id,
            runs=runs_by_id,
            operations=operations_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        upstream_dataset_ids = {input_.dataset_id for input_ in inputs} - ids_to_skip.datasets
        downstream_dataset_ids = {output.dataset_id for output in outputs} - ids_to_skip.datasets
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `dataset1 -> current_operation -> dataset2``,
            # but do not include relations like `another_operation -> dataset2`.
            # Also datasets and symlinks will be populated by nested calls.
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=upstream_dataset_ids,
                    direction=LineageDirectionV1.UPSTREAM,
                    granularity="OPERATION",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_datasets(
                    start_node_ids=downstream_dataset_ids,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    granularity="OPERATION",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            dataset_ids = upstream_dataset_ids | downstream_dataset_ids
            datasets = await self._uow.dataset.list_by_ids(dataset_ids)

            extra_datasets, dataset_symlinks = await self._extend_datasets_with_symlinks(
                dataset_ids,
                ids_to_skip,
            )
            datasets.extend(extra_datasets)

            result.datasets = {dataset.id: dataset for dataset in datasets}
            result.dataset_symlinks = {
                (dataset_symlink.from_dataset_id, dataset_symlink.to_dataset_id): dataset_symlink
                for dataset_symlink in dataset_symlinks
            }

        if level == 0:
            if include_column_lineage:
                result.column_lineage.update(
                    await self._get_column_lineage(
                        current_result=result,
                        since=since,
                        until=until,
                        granularity="OPERATION",
                    ),
                )

            result = self._drop_unused_datasets(result)

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Found %d jobs, %d runs, %d operations, %d datasets, %d dataset symlinks, "
                "%d inputs, %d outputs, %d column lineage",
                level,
                len(result.jobs),
                len(result.runs),
                len(result.operations),
                len(result.datasets),
                len(result.dataset_symlinks),
                len(result.inputs),
                len(result.outputs),
                len(result.column_lineage),
            )
        return result

    async def get_lineage_by_datasets(
        self,
        start_node_ids: Collection[int],
        direction: LineageDirectionV1,
        granularity: Literal["JOB", "RUN", "OPERATION"],
        since: datetime,
        until: datetime | None,
        depth: int,
        include_column_lineage: bool = False,  # noqa: FBT001, FBT002
        ids_to_skip: IdsToSkip | None = None,
        level: int = 0,
    ) -> LineageServiceResult:
        if not start_node_ids:
            return LineageServiceResult()

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Get lineage by datasets %r, with direction %s, since %s, until %s",
                level,
                sorted(start_node_ids),
                direction,
                since,
                until,
            )

        datasets = await self._uow.dataset.list_by_ids(start_node_ids)
        if not datasets:
            logger.info("[Level %d] No datasets found", level)
            return LineageServiceResult()

        datasets_by_id = {dataset.id: dataset for dataset in datasets}

        # Threat dataset symlinks like they are specified in `start_node_ids`
        ids_to_skip = ids_to_skip or IdsToSkip()
        extra_datasets, dataset_symlinks = await self._extend_datasets_with_symlinks(datasets_by_id, ids_to_skip)
        datasets.extend(extra_datasets)

        datasets_by_id.update({dataset.id: dataset for dataset in extra_datasets})
        dataset_symlinks_by_id = {
            (dataset_symlink.from_dataset_id, dataset_symlink.to_dataset_id): dataset_symlink
            for dataset_symlink in dataset_symlinks
        }

        # Get datasets -> (inputs, outputs) -> operations -> runs -> jobs
        match granularity:
            case "OPERATION":
                result = await self._dataset_lineage_with_operation_granularity(
                    datasets_by_id=datasets_by_id,
                    dataset_symlinks_by_id=dataset_symlinks_by_id,
                    direction=direction,
                    since=since,
                    until=until,
                    depth=depth,
                    level=level,
                    ids_to_skip=ids_to_skip,
                )
            case "RUN":
                result = await self._dataset_lineage_with_run_granularity(
                    datasets_by_id=datasets_by_id,
                    dataset_symlinks_by_id=dataset_symlinks_by_id,
                    direction=direction,
                    since=since,
                    until=until,
                    depth=depth,
                    level=level,
                    ids_to_skip=ids_to_skip,
                )
            case "JOB":
                result = await self._dataset_lineage_with_job_granularity(
                    datasets_by_id=datasets_by_id,
                    dataset_symlinks_by_id=dataset_symlinks_by_id,
                    direction=direction,
                    since=since,
                    until=until,
                    depth=depth,
                    level=level,
                    ids_to_skip=ids_to_skip,
                )
            case _:
                msg = f"Unknown granularity: {granularity}"
                raise ValueError(msg)

        if level == 0:
            if include_column_lineage:
                result.column_lineage.update(
                    await self._get_column_lineage(
                        current_result=result,
                        since=since,
                        until=until,
                        granularity=granularity,
                    ),
                )

            result = self._drop_unused_datasets(result, keep_dataset_ids=start_node_ids)

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "[Level %d] Found %d jobs, %d runs, %d operations, %d datasets, %d dataset symlinks, "
                "%d inputs, %d outputs, %d column lineage",
                level,
                len(result.jobs),
                len(result.runs),
                len(result.operations),
                len(result.datasets),
                len(result.dataset_symlinks),
                len(result.inputs),
                len(result.outputs),
                len(result.column_lineage),
            )
        return result

    async def _extend_datasets_with_symlinks(
        self,
        dataset_ids: Collection[int],
        ids_to_skip: IdsToSkip | None = None,
    ) -> tuple[list[Dataset], list[DatasetSymlink]]:
        # For now return all symlinks regardless of direction
        dataset_symlinks = await self._uow.dataset_symlink.list_by_dataset_ids(dataset_ids)

        ids_to_skip = ids_to_skip or IdsToSkip()
        dataset_ids_from_symlinks = {dataset_symlink.from_dataset_id for dataset_symlink in dataset_symlinks}
        dataset_ids_to_symlinks = {dataset_symlink.to_dataset_id for dataset_symlink in dataset_symlinks}
        dataset_ids_to_load = (
            (dataset_ids_from_symlinks | dataset_ids_to_symlinks) - set(dataset_ids) - ids_to_skip.datasets
        )

        datasets_from_symlinks = await self._uow.dataset.list_by_ids(dataset_ids_to_load)
        return datasets_from_symlinks, dataset_symlinks

    async def _dataset_lineage_with_operation_granularity(
        self,
        datasets_by_id: dict[int, Dataset],
        dataset_symlinks_by_id: dict[tuple[int, int], DatasetSymlink],
        direction: LineageDirectionV1,
        since: datetime,
        until: datetime | None,
        depth: int,
        level: int,
        ids_to_skip: IdsToSkip | None = None,
    ) -> LineageServiceResult:
        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            inputs.extend(
                await self._uow.input.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="OPERATION",
                ),
            )
            # include sum over all operations
            inputs.extend(
                await self._uow.input.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="RUN",
                ),
            )
            # include sum over all runs
            inputs.extend(
                await self._uow.input.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            outputs.extend(
                await self._uow.output.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="OPERATION",
                ),
            )
            # include sum over all operations
            outputs.extend(
                await self._uow.output.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="RUN",
                ),
            )
            # include sum over all runs
            outputs.extend(
                await self._uow.output.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}

        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        result = LineageServiceResult(
            datasets=datasets_by_id,
            dataset_symlinks=dataset_symlinks_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        ids_to_skip = ids_to_skip or IdsToSkip()
        downstream_operation_ids = {input_.operation_id for input_ in inputs if input_.operation_id is not None}
        upstream_operation_ids = {output.operation_id for output in outputs if output.operation_id is not None}

        # ignore operations only after they were included
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `operation1 -> current_dataset -> operation2``,
            # but do not include relations like `another_dataset -> operation2`.
            # Also operations, runs and jobs will be populated by nested call.
            result.merge(
                await self.get_lineage_by_operations(
                    start_node_ids=downstream_operation_ids - ids_to_skip.operations,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_operations(
                    start_node_ids=upstream_operation_ids - ids_to_skip.operations,
                    direction=LineageDirectionV1.UPSTREAM,
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            operation_ids = downstream_operation_ids | upstream_operation_ids - ids_to_skip.operations
            operations = await self._uow.operation.list_by_ids(operation_ids)
            result.operations = {operation.id: operation for operation in operations}

            run_ids = {operation.run_id for operation in operations}
            runs = await self._uow.run.list_by_ids(run_ids - ids_to_skip.runs)
            result.runs = {run.id: run for run in runs}

            job_ids = {run.job_id for run in runs}
            jobs = await self._uow.job.list_by_ids(job_ids - ids_to_skip.jobs)
            result.jobs = {job.id: job for job in jobs}

        return result

    async def _dataset_lineage_with_run_granularity(
        self,
        datasets_by_id: dict[int, Dataset],
        dataset_symlinks_by_id: dict[tuple[int, int], DatasetSymlink],
        direction: LineageDirectionV1,
        since: datetime,
        until: datetime | None,
        depth: int,
        level: int,
        ids_to_skip: IdsToSkip | None = None,
    ) -> LineageServiceResult:
        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            inputs.extend(
                await self._uow.input.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="RUN",
                ),
            )
            # include sum over all runs
            inputs.extend(
                await self._uow.input.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            outputs.extend(
                await self._uow.output.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="RUN",
                ),
            )
            # include sum over all runs
            outputs.extend(
                await self._uow.output.list_by_dataset_ids(
                    datasets_by_id,
                    since=since,
                    until=until,
                    granularity="JOB",
                ),
            )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}
        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        result = LineageServiceResult(
            datasets=datasets_by_id,
            dataset_symlinks=dataset_symlinks_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        ids_to_skip = ids_to_skip or IdsToSkip()
        downstream_run_ids = {input_.run_id for input_ in inputs if input_.run_id is not None}
        upstream_run_ids = {output.run_id for output in outputs if output.run_id is not None}
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `run1 -> current_dataset -> run2``,
            # but do not include relations like `another_dataset -> run2`.
            # Also runs and jobs will be populated by nested call.
            result.merge(
                await self.get_lineage_by_runs(
                    start_node_ids=downstream_run_ids - ids_to_skip.runs,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    granularity="RUN",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_runs(
                    start_node_ids=upstream_run_ids - ids_to_skip.runs,
                    direction=LineageDirectionV1.UPSTREAM,
                    granularity="RUN",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            run_ids = downstream_run_ids | upstream_run_ids - ids_to_skip.runs
            runs = await self._uow.run.list_by_ids(run_ids)
            result.runs = {run.id: run for run in runs}

            job_ids = {run.job_id for run in runs}
            jobs = await self._uow.job.list_by_ids(job_ids - ids_to_skip.jobs)
            result.jobs = {job.id: job for job in jobs}

        return result

    async def _dataset_lineage_with_job_granularity(
        self,
        datasets_by_id: dict[int, Dataset],
        dataset_symlinks_by_id: dict[tuple[int, int], DatasetSymlink],
        direction: LineageDirectionV1,
        since: datetime,
        until: datetime | None,
        depth: int,
        level: int,
        ids_to_skip: IdsToSkip | None = None,
    ) -> LineageServiceResult:
        inputs: list[InputRow] = []
        outputs: list[OutputRow] = []
        if direction in {LineageDirectionV1.DOWNSTREAM, LineageDirectionV1.BOTH}:
            inputs = await self._uow.input.list_by_dataset_ids(
                datasets_by_id,
                since=since,
                until=until,
                granularity="JOB",
            )

        if direction in {LineageDirectionV1.UPSTREAM, LineageDirectionV1.BOTH}:
            outputs = await self._uow.output.list_by_dataset_ids(
                datasets_by_id,
                since=since,
                until=until,
                granularity="JOB",
            )

        input_schema_ids = {input_.schema_id for input_ in inputs if input_.schema_id is not None}
        output_schema_ids = {output.schema_id for output in outputs if output.schema_id is not None}
        schemas = await self._uow.schema.list_by_ids(input_schema_ids | output_schema_ids)
        schemas_by_id = {schema.id: schema for schema in schemas}

        for input_ in inputs:
            if input_.schema_id is not None:
                input_.schema = schemas_by_id.get(input_.schema_id)
        for output in outputs:
            if output.schema_id is not None:
                output.schema = schemas_by_id.get(output.schema_id)

        result = LineageServiceResult(
            datasets=datasets_by_id,
            dataset_symlinks=dataset_symlinks_by_id,
            inputs={
                (input_.dataset_id, input_.job_id, input_.run_id, input_.operation_id): input_ for input_ in inputs
            },
            outputs={
                (output.dataset_id, output.job_id, output.run_id, output.operation_id, output.type): output
                for output in outputs
            },
        )

        ids_to_skip = ids_to_skip or IdsToSkip()
        downstream_job_ids = {input_.job_id for input_ in inputs if input_.job_id is not None}
        upstream_job_ids = {output.job_id for output in outputs if output.job_id is not None}
        ids_to_skip = ids_to_skip.merge(IdsToSkip.from_result(result))

        if depth > 1:
            # If we passed direction=BOTH, return only relations like `job1 -> current_dataset -> job2``,
            # but do not include relations like `another_dataset -> job2`.
            # Also jobs will be populated by nested call.
            result.merge(
                await self.get_lineage_by_jobs(
                    start_node_ids=downstream_job_ids - ids_to_skip.jobs,
                    direction=LineageDirectionV1.DOWNSTREAM,
                    granularity="JOB",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
            result.merge(
                await self.get_lineage_by_jobs(
                    start_node_ids=upstream_job_ids - ids_to_skip.jobs,
                    direction=LineageDirectionV1.UPSTREAM,
                    granularity="JOB",
                    since=since,
                    until=until,
                    depth=depth - 1,
                    level=level + 1,
                    ids_to_skip=ids_to_skip,
                ),
            )
        else:
            job_ids = downstream_job_ids | upstream_job_ids - ids_to_skip.jobs
            jobs = await self._uow.job.list_by_ids(job_ids)
            result.jobs = {job.id: job for job in jobs}

        return result

    async def _get_column_lineage(
        self,
        current_result: LineageServiceResult,
        since: datetime,
        until: datetime | None,
        granularity: Literal["OPERATION", "RUN", "JOB"],
    ) -> dict[tuple[int, int], list[ColumnLineageRow]]:
        """
        'granularity' argument of this function not the same as granularity in api request.
        Here it's used for aggregation entity
        """
        result: dict[tuple[int, int], list[ColumnLineageRow]] = defaultdict(list)
        if not current_result.inputs or not current_result.outputs:
            return result

        source_dataset_ids = [input_.dataset_id for input_ in current_result.inputs.values()]
        target_dataset_ids = [output.dataset_id for output in current_result.outputs.values()]

        match granularity:
            case "OPERATION":
                column_lineage_result = await self._uow.column_lineage.list_by_operation_ids(
                    operation_ids=current_result.operations,
                    # return column lineage only for datasets included into response
                    source_dataset_ids=source_dataset_ids,
                    target_dataset_ids=target_dataset_ids,
                )
            case "RUN":
                column_lineage_result = await self._uow.column_lineage.list_by_run_ids(
                    run_ids=current_result.runs,
                    since=since,
                    until=until,
                    source_dataset_ids=source_dataset_ids,
                    target_dataset_ids=target_dataset_ids,
                )
            case "JOB":
                column_lineage_result = await self._uow.column_lineage.list_by_job_ids(
                    job_ids=current_result.jobs,
                    since=since,
                    until=until,
                    source_dataset_ids=source_dataset_ids,
                    target_dataset_ids=target_dataset_ids,
                )
            case _:
                msg = f"Unknown granularity for column lineage: {granularity}"
                raise ValueError(msg)

        for relation in column_lineage_result:
            result[(relation.source_dataset_id, relation.target_dataset_id)].append(
                relation,
            )

        return result

    def _drop_unused_datasets(
        self,
        current_result: LineageServiceResult,
        keep_dataset_ids: Collection[int] | None = None,
    ) -> LineageServiceResult:
        dataset_ids_with_inputs = {input_.dataset_id for input_ in current_result.inputs.values()}
        dataset_ids_with_outputs = {output.dataset_id for output in current_result.outputs.values()}

        keep_ids = set(keep_dataset_ids or []) | dataset_ids_with_inputs | dataset_ids_with_outputs

        final_datasets = {key: dataset for key, dataset in current_result.datasets.items() if dataset.id in keep_ids}
        final_symlinks = {
            key: symlink
            for key, symlink in current_result.dataset_symlinks.items()
            if symlink.from_dataset_id in keep_ids and symlink.to_dataset_id in keep_ids
        }

        return dataclasses.replace(
            current_result,
            datasets=final_datasets,
            dataset_symlinks=final_symlinks,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from fastapi import FastAPI
from starlette.middleware.sessions import SessionMiddleware

from data_rentgen.server.settings.session import SessionSettings


def apply_session_middleware(app: FastAPI, settings: SessionSettings) -> FastAPI:
    """Add SessionMiddleware middleware to the application."""

    app.add_middleware(
        SessionMiddleware,
        **settings.model_dump(),
    )
    return app



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from fastapi import Query
from pydantic import BaseModel, ConfigDict, Field

from data_rentgen.server.schemas.v1.address import AddressResponseV1
from data_rentgen.server.schemas.v1.pagination import PaginateQueryV1


class LocationResponseV1(BaseModel):
    id: str = Field(description="Location id", coerce_numbers_to_str=True)
    type: str = Field(description="Location type, e.g kafka, hdfs, postgres")
    name: str = Field(description="Location name, e.g. cluster name")
    addresses: list[AddressResponseV1] = Field(description="List of addresses")
    external_id: str | None = Field(description="External ID for integration with other systems")

    model_config = ConfigDict(from_attributes=True)


class LocationDatasetStatisticsReponseV1(BaseModel):
    """Location dataset statistics response."""

    total_datasets: int = Field(description="Total number of datasets bound to this location")

    model_config = ConfigDict(from_attributes=True)


class LocationJobStatisticsReponseV1(BaseModel):
    """Location job statistics response."""

    total_jobs: int = Field(description="Total number of jobs bound to this location")

    model_config = ConfigDict(from_attributes=True)


class LocationStatisticsReponseV1(BaseModel):
    """Location statistics response."""

    datasets: LocationDatasetStatisticsReponseV1 = Field(description="Dataset statistics")
    jobs: LocationJobStatisticsReponseV1 = Field(description="Dataset statistics")

    model_config = ConfigDict(from_attributes=True)


class LocationDetailedResponseV1(BaseModel):
    id: str = Field(description="Location id", coerce_numbers_to_str=True)
    data: LocationResponseV1 = Field(description="Location data")
    statistics: LocationStatisticsReponseV1 = Field(description="Location statistics")

    model_config = ConfigDict(from_attributes=True)


class LocationPaginateQueryV1(PaginateQueryV1):
    """Query params for Location paginate request."""

    location_id: list[int] = Field(Query(default_factory=list), description="Location id")
    location_type: str | None = Field(Query(default=None), description="Location type")
    search_query: str | None = Field(
        Query(
            default=None,
            min_length=3,
            description="Search query",
        ),
    )

    model_config = ConfigDict(extra="forbid")


class UpdateLocationRequestV1(BaseModel):
    external_id: str | None = Field(description="External ID for integration with other systems")

    model_config = ConfigDict(extra="forbid")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from data_rentgen.dto.location import LocationDTO


class JobTypeDTO(str, Enum):
    AIRFLOW_DAG = "AIRFLOW_DAG"
    AIRFLOW_TASK = "AIRFLOW_TASK"
    SPARK_APPLICATION = "SPARK_APPLICATION"

    def __str__(self) -> str:
        return self.value


@dataclass
class JobDTO:
    name: str
    location: LocationDTO
    type: JobTypeDTO | None = None
    id: int | None = field(default=None, compare=False)

    @property
    def unique_key(self) -> tuple:
        return (self.location.unique_key, self.name)

    def merge(self, new: JobDTO) -> JobDTO:
        if new.id is None and self.type == new.type:
            # jobs aren't changed that much, reuse them if possible
            return self

        return JobDTO(
            location=self.location.merge(new.location),
            name=self.name,
            type=new.type or self.type,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from enum import Enum

from sqlalchemy import BigInteger, ForeignKey, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy_utils import ChoiceType

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.dataset import Dataset


class DatasetSymlinkType(str, Enum):
    METASTORE = "METASTORE"
    WAREHOUSE = "WAREHOUSE"

    def __str__(self) -> str:
        return self.value


class DatasetSymlink(Base):
    __tablename__ = "dataset_symlink"
    __table_args__ = (UniqueConstraint("from_dataset_id", "to_dataset_id"),)

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    from_dataset_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("dataset.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    from_dataset: Mapped[Dataset] = relationship(
        Dataset,
        lazy="noload",
        foreign_keys=[from_dataset_id],
    )

    to_dataset_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("dataset.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    to_dataset: Mapped[Dataset] = relationship(
        Dataset,
        lazy="noload",
        foreign_keys=[to_dataset_id],
    )

    type: Mapped[DatasetSymlinkType] = mapped_column(
        ChoiceType(DatasetSymlinkType, impl=String(32)),
        nullable=False,
        doc="Type of dataset symlink, e.g. metastore table -> hdfs location",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet


class OpenLineageSparkJobDetailsRunFacet(OpenLineageRunFacet):
    """Run facet describing Spark job.
    See [SparkJobDetailsFacet](https://github.com/OpenLineage/OpenLineage/blob/main/integration/spark/shared/src/main/java/io/openlineage/spark/agent/facets/SparkJobDetailsFacet.java).
    """

    jobId: int
    jobDescription: str | None = None
    jobGroup: str | None = None
    jobCallSite: str | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Literal
from uuid import UUID

from sqlalchemy import ColumnElement, Row, any_, func, literal_column, select
from sqlalchemy.dialects.postgresql import insert

from data_rentgen.db.models import Output, OutputType, Schema
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.uuid import (
    extract_timestamp_from_uuid,
    generate_incremental_uuid,
)
from data_rentgen.dto import OutputDTO


@dataclass
class OutputRow:
    created_at: datetime
    operation_id: UUID
    run_id: UUID
    job_id: int
    dataset_id: int
    num_bytes: int | None
    num_rows: int | None
    num_files: int | None
    type: str | None = None
    schema_id: int | None = None
    schema_relevance_type: Literal["EXACT_MATCH", "LATEST_KNOWN"] | None = None
    schema: Schema | None = None


class OutputRepository(Repository[Output]):
    def get_id(self, output: OutputDTO) -> UUID:
        # `created_at' field of output should be the same as operation's,
        # to avoid scanning all partitions and speed up queries
        created_at = extract_timestamp_from_uuid(output.operation.id)

        # instead of using UniqueConstraint on multiple fields, one of which (schema_id) can be NULL,
        # use them to calculate unique id
        id_components = [
            str(output.operation.id),
            str(output.dataset.id),
            str(output.schema.id) if output.schema else "",
        ]
        return generate_incremental_uuid(created_at, ".".join(id_components))

    async def create_or_update_bulk(self, outputs: list[OutputDTO]) -> None:
        if not outputs:
            return

        insert_statement = insert(Output)
        new_row = insert_statement.excluded
        statement = insert_statement.on_conflict_do_update(
            index_elements=[Output.created_at, Output.id],
            set_={
                "num_bytes": func.greatest(new_row.num_bytes, Output.num_bytes),
                "num_rows": func.greatest(new_row.num_rows, Output.num_rows),
                "num_files": func.greatest(new_row.num_files, Output.num_files),
            },
        )

        await self._session.execute(
            statement,
            [
                {
                    "id": self.get_id(output),
                    "created_at": extract_timestamp_from_uuid(output.operation.id),
                    "type": OutputType(output.type),
                    "operation_id": output.operation.id,
                    "run_id": output.operation.run.id,
                    "job_id": output.operation.run.job.id,  # type: ignore[arg-type]
                    "dataset_id": output.dataset.id,  # type: ignore[arg-type]
                    "schema_id": output.schema.id if output.schema else None,
                    "num_bytes": output.num_bytes,
                    "num_rows": output.num_rows,
                    "num_files": output.num_files,
                }
                for output in outputs
            ],
        )

    async def list_by_operation_ids(
        self,
        operation_ids: Collection[UUID],
        granularity: Literal["JOB", "RUN", "OPERATION"],
    ) -> list[OutputRow]:
        if not operation_ids:
            return []

        # Output created_at is always the same as operation's created_at
        # do not use `tuple_(Output.created_at, Output.operation_id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        min_created_at = extract_timestamp_from_uuid(min(operation_ids))
        max_created_at = extract_timestamp_from_uuid(max(operation_ids))
        where = [
            Output.created_at >= min_created_at,
            Output.created_at <= max_created_at,
            Output.operation_id == any_(list(operation_ids)),  # type: ignore[arg-type]
        ]

        return await self._get_outputs(where, granularity)

    async def list_by_run_ids(
        self,
        run_ids: Collection[UUID],
        since: datetime,
        until: datetime | None,
        granularity: Literal["JOB", "RUN", "OPERATION"],
    ) -> list[OutputRow]:
        if not run_ids:
            return []

        min_run_created_at = extract_timestamp_from_uuid(min(run_ids))
        min_created_at = max(min_run_created_at, since.astimezone(timezone.utc))

        where = [
            Output.created_at >= min_created_at,
            Output.run_id == any_(list(run_ids)),  # type: ignore[arg-type]
        ]
        if until:
            where.append(Output.created_at <= until)

        return await self._get_outputs(where, granularity)

    async def list_by_job_ids(
        self,
        job_ids: Collection[int],
        since: datetime,
        until: datetime | None,
        granularity: Literal["JOB", "RUN", "OPERATION"],
    ) -> list[OutputRow]:
        if not job_ids:
            return []

        where = [
            Output.created_at >= since,
            Output.job_id == any_(list(job_ids)),  # type: ignore[arg-type]
        ]
        if until:
            where.append(Output.created_at <= until)

        return await self._get_outputs(where, granularity)

    async def list_by_dataset_ids(
        self,
        dataset_ids: Collection[int],
        since: datetime,
        until: datetime | None,
        granularity: Literal["JOB", "RUN", "OPERATION"],
    ) -> list[OutputRow]:
        if not dataset_ids:
            return []

        where = [
            Output.created_at >= since,
            Output.dataset_id == any_(list(dataset_ids)),  # type: ignore[arg-type]
        ]
        if until:
            where.append(Output.created_at <= until)

        return await self._get_outputs(where, granularity)

    async def _get_outputs(
        self,
        where: Collection[ColumnElement],
        granularity: Literal["JOB", "RUN", "OPERATION"],
    ) -> list[OutputRow]:
        if granularity == "OPERATION":
            # return Output as-is
            simple_query = select(Output).where(*where)
            result = await self._session.scalars(simple_query)
            return [
                OutputRow(
                    created_at=row.created_at,
                    operation_id=row.operation_id,
                    run_id=row.run_id,
                    job_id=row.job_id,
                    dataset_id=row.dataset_id,
                    num_bytes=row.num_bytes,
                    num_rows=row.num_rows,
                    num_files=row.num_files,
                    schema_id=row.schema_id,
                    schema_relevance_type="EXACT_MATCH" if row.schema_id else None,
                    type=row.type,
                )
                for row in result.all()
            ]

        # return an aggregated Output
        if granularity == "RUN":
            partition_by = [Output.run_id, Output.job_id, Output.dataset_id]
            base_query = (
                select(
                    Output,
                    func.first_value(Output.schema_id)
                    .over(partition_by=partition_by, order_by=[Output.created_at, Output.schema_id])
                    .label("oldest_schema_id"),
                    func.last_value(Output.schema_id)
                    .over(partition_by=partition_by, order_by=[Output.created_at, Output.schema_id])
                    .label("newest_schema_id"),
                )
                .where(*where)
                .cte()
            )

            query = select(
                func.max(base_query.c.created_at).label("created_at"),
                literal_column("NULL").label("id"),
                literal_column("NULL").label("operation_id"),
                base_query.c.run_id,
                base_query.c.job_id,
                base_query.c.dataset_id,
                func.sum(base_query.c.num_bytes).label("sum_num_bytes"),
                func.sum(base_query.c.num_rows).label("sum_num_rows"),
                func.sum(base_query.c.num_files).label("sum_num_files"),
                func.min(base_query.c.type).label("min_type"),
                func.max(base_query.c.type).label("max_type"),
                func.min(base_query.c.oldest_schema_id).label("min_schema_id"),
                func.max(base_query.c.newest_schema_id).label("max_schema_id"),
            ).group_by(
                base_query.c.run_id,
                base_query.c.job_id,
                base_query.c.dataset_id,
            )
        else:
            partition_by = [Output.job_id, Output.dataset_id]
            base_query = (
                select(
                    Output,
                    func.first_value(Output.schema_id)
                    .over(partition_by=partition_by, order_by=[Output.created_at, Output.schema_id])
                    .label("oldest_schema_id"),
                    func.last_value(Output.schema_id)
                    .over(partition_by=partition_by, order_by=[Output.created_at, Output.schema_id])
                    .label("newest_schema_id"),
                )
                .where(*where)
                .cte()
            )

            query = select(
                func.max(base_query.c.created_at).label("created_at"),
                literal_column("NULL").label("id"),
                literal_column("NULL").label("operation_id"),
                literal_column("NULL").label("run_id"),
                base_query.c.job_id,
                base_query.c.dataset_id,
                func.sum(base_query.c.num_bytes).label("sum_num_bytes"),
                func.sum(base_query.c.num_rows).label("sum_num_rows"),
                func.sum(base_query.c.num_files).label("sum_num_files"),
                func.min(base_query.c.type).label("min_type"),
                func.max(base_query.c.type).label("max_type"),
                func.min(base_query.c.oldest_schema_id).label("min_schema_id"),
                func.max(base_query.c.newest_schema_id).label("max_schema_id"),
            ).group_by(
                base_query.c.job_id,
                base_query.c.dataset_id,
            )
        query_result = await self._session.execute(query)
        results = []
        for row in query_result.all():
            schema_relevance_type: Literal["EXACT_MATCH", "LATEST_KNOWN"] | None
            if row.max_schema_id:
                schema_relevance_type = "EXACT_MATCH" if row.min_schema_id == row.max_schema_id else "LATEST_KNOWN"
            else:
                schema_relevance_type = None
            results.append(
                OutputRow(
                    created_at=row.created_at,
                    operation_id=row.operation_id,
                    run_id=row.run_id,
                    job_id=row.job_id,
                    dataset_id=row.dataset_id,
                    num_bytes=row.sum_num_bytes,
                    num_rows=row.sum_num_rows,
                    num_files=row.sum_num_files,
                    schema_id=row.max_schema_id,
                    schema_relevance_type=schema_relevance_type,
                    # If all outputs within Dataset -> Run|Job have the same type, save it.
                    # If not, it's impossible to merge.
                    type=row.max_type if row.min_type == row.max_type else None,
                ),
            )
        return results

    async def get_stats_by_operation_ids(self, operation_ids: Collection[UUID]) -> dict[UUID, Row]:
        if not operation_ids:
            return {}

        # Input created_at is always the same as operation's created_at
        # do not use `tuple_(Input.created_at, Input.operation_id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        min_created_at = extract_timestamp_from_uuid(min(operation_ids))
        max_created_at = extract_timestamp_from_uuid(max(operation_ids))

        query = (
            select(
                Output.operation_id.label("operation_id"),
                func.count(Output.dataset_id.distinct()).label("total_datasets"),
                func.sum(Output.num_bytes).label("total_bytes"),
                func.sum(Output.num_rows).label("total_rows"),
                func.sum(Output.num_files).label("total_files"),
            )
            .where(
                Output.created_at >= min_created_at,
                Output.created_at <= max_created_at,
                Output.operation_id == any_(list(operation_ids)),  # type: ignore[arg-type]
            )
            .group_by(Output.operation_id)
        )

        query_result = await self._session.execute(query)
        return {row.operation_id: row for row in query_result.all()}

    async def get_stats_by_run_ids(self, run_ids: Collection[UUID]) -> dict[UUID, Row]:
        if not run_ids:
            return {}

        # unlike list_by_run_ids, we need to get all statistics for specific runs, regardless of time range
        min_created_at = extract_timestamp_from_uuid(min(run_ids))
        query = (
            select(
                Output.run_id.label("run_id"),
                func.count(Output.dataset_id.distinct()).label("total_datasets"),
                func.sum(Output.num_bytes).label("total_bytes"),
                func.sum(Output.num_rows).label("total_rows"),
                func.sum(Output.num_files).label("total_files"),
            )
            .where(
                Output.created_at >= min_created_at,
                Output.run_id == any_(list(run_ids)),  # type: ignore[arg-type]
            )
            .group_by(Output.run_id)
        )

        query_result = await self._session.execute(query)
        return {row.run_id: row for row in query_result.all()}



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.extractors.dataset import extract_dataset_and_symlinks
from data_rentgen.consumer.extractors.schema import extract_schema
from data_rentgen.consumer.openlineage.dataset import OpenLineageOutputDataset
from data_rentgen.dto import DatasetSymlinkDTO, OutputDTO, OutputTypeDTO
from data_rentgen.dto.operation import OperationDTO


def extract_output(
    operation: OperationDTO,
    dataset: OpenLineageOutputDataset,
) -> tuple[OutputDTO, list[DatasetSymlinkDTO]]:
    lifecycle_change = dataset.facets.lifecycleStateChange
    if lifecycle_change:  # noqa: SIM108
        output_type = OutputTypeDTO(lifecycle_change.lifecycleStateChange)
    else:
        output_type = OutputTypeDTO.APPEND
    dataset_dto, symlinks = extract_dataset_and_symlinks(dataset)
    result = OutputDTO(
        type=output_type,
        operation=operation,
        dataset=dataset_dto,
        schema=extract_schema(dataset),
    )
    if dataset.outputFacets.outputStatistics:
        result.num_rows = dataset.outputFacets.outputStatistics.rows
        result.num_bytes = dataset.outputFacets.outputStatistics.bytes
        result.num_files = dataset.outputFacets.outputStatistics.files

    return result, symlinks



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import datetime
from enum import IntEnum
from uuid import UUID

from fastapi import Query
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)

from data_rentgen.server.schemas.v1.pagination import PaginateQueryV1
from data_rentgen.utils import UUIDv6Plus


class OperationStatusV1(IntEnum):
    UNKNOWN = -1
    """No data about status"""

    STARTED = 0
    """Received START event"""

    SUCCEEDED = 1
    """Finished successfully"""

    FAILED = 2
    """Internal failure"""

    KILLED = 3
    """Killed externally, e.g. by user request or in case of OOM"""

    def __str__(self) -> str:
        return self.name


class OperationResponseV1(BaseModel):
    """Operation response."""

    id: UUID = Field(description="Operation id")
    created_at: datetime = Field(description="Operation creation time")
    run_id: UUID = Field(description="Run operation belongs to")
    status: OperationStatusV1 = Field(description="Operation status")
    name: str = Field(description="Operation name")
    type: str = Field(description="Operation type")

    position: int | None = Field(description="Sequentinal position of operation within the run, e.g. Spark jobId")
    group: str | None = Field(description="Operation group, e.g. Spark jobGroup")
    description: str | None = Field(description="Operation description, e.g. Spark jobDescription")

    started_at: datetime | None = Field(description="Start time of the Operation")
    ended_at: datetime | None = Field(description="End time of the Operation")

    model_config = ConfigDict(from_attributes=True)

    @field_serializer("status", when_used="json-unless-none")
    def _serialize_status(self, value: OperationStatusV1) -> str:
        return str(value)


class OperationIOStatisticsReponseV1(BaseModel):
    """Operation IO statistics response."""

    total_datasets: int = Field(default=0, description="Total number of datasets")
    total_bytes: int = Field(default=0, description="Total number of bytes")
    total_rows: int = Field(default=0, description="Total number of rows")
    total_files: int = Field(default=0, description="Total number of files")

    model_config = ConfigDict(from_attributes=True)


class OperationStatisticsReponseV1(BaseModel):
    """Operation statistics response."""

    outputs: OperationIOStatisticsReponseV1 = Field(description="Output statistics")
    inputs: OperationIOStatisticsReponseV1 = Field(description="Input statistics")

    model_config = ConfigDict(from_attributes=True)


class OperationDetailedResponseV1(BaseModel):
    """Operation response."""

    id: UUID = Field(description="Operation id")
    data: OperationResponseV1 = Field(description="Operation data")
    statistics: OperationStatisticsReponseV1 = Field(
        description="Operation statistics",
    )

    model_config = ConfigDict(from_attributes=True)


class OperationQueryV1(PaginateQueryV1):
    """Query params for Operations paginate request."""

    since: datetime | None = Field(
        Query(
            default=None,
            description="Minimum value of Operation 'created_at' field, in ISO 8601 format",
            examples=["2008-09-15T15:53:00+05:00"],
        ),
    )
    until: datetime | None = Field(
        Query(
            default=None,
            description="Maximum value of Operation 'created_at' field, in ISO 8601 format",
            examples=["2008-09-15T15:53:00+05:00"],
        ),
    )
    operation_id: list[UUIDv6Plus] = Field(
        Query(
            default_factory=list,
            description="Operation ids, for exact match",
        ),
    )
    run_id: UUIDv6Plus | None = Field(
        Query(
            default=None,
            description="Run id, can be used only with 'since'",
        ),
    )

    model_config = ConfigDict(extra="forbid")

    @field_validator("until", mode="after")
    @classmethod
    def _check_until(cls, value: datetime | None, info: ValidationInfo) -> datetime | None:
        since = info.data.get("since")
        if since and value and since >= value:
            msg = "'since' should be less than 'until'"
            raise ValueError(msg)
        return value

    @model_validator(mode="after")
    def _check_fields(self):
        if not any([self.operation_id, self.run_id]):
            msg = "input should contain either 'run_id' or 'operation_id' field"
            raise ValueError(msg)
        if self.run_id and not self.since:
            msg = "'run_id' can be passed only with 'since'"
            raise ValueError(msg)
        return self



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.extractors.dataset import extract_dataset_and_symlinks
from data_rentgen.consumer.extractors.schema import extract_schema
from data_rentgen.consumer.openlineage.dataset import OpenLineageInputDataset
from data_rentgen.dto import DatasetSymlinkDTO, InputDTO
from data_rentgen.dto.operation import OperationDTO


def extract_input(
    operation: OperationDTO,
    dataset: OpenLineageInputDataset,
) -> tuple[InputDTO, list[DatasetSymlinkDTO]]:
    dataset_dto, symlinks = extract_dataset_and_symlinks(dataset)
    result = InputDTO(
        operation=operation,
        dataset=dataset_dto,
        schema=extract_schema(dataset),
    )
    if dataset.inputFacets.inputStatistics:
        result.num_rows = dataset.inputFacets.inputStatistics.rows
        result.num_bytes = dataset.inputFacets.inputStatistics.bytes
        result.num_files = dataset.inputFacets.inputStatistics.files
    return result, symlinks



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from enum import Enum

from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet


class OpenLineageSparkDeployMode(str, Enum):
    """Spark application deploy mode."""

    CLIENT = "client"
    CLUSTER = "cluster"

    def __str__(self) -> str:
        return self.value


class OpenLineageSparkApplicationDetailsRunFacet(OpenLineageRunFacet):
    """Run facet describing Spark application.
    See [SparkApplicationDetailsFacet](https://github.com/OpenLineage/OpenLineage/blob/main/integration/spark/shared/src/main/java/io/openlineage/spark/agent/facets/SparkApplicationDetailsFacet.java).
    """

    master: str
    appName: str
    applicationId: str
    deployMode: OpenLineageSparkDeployMode
    driverHost: str
    userName: str
    uiWebUrl: str | None = None
    proxyUrl: str | None = None
    historyUrl: str | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from uuid import UUID

from sqlalchemy import JSON, BigInteger
from sqlalchemy import UUID as SQL_UUID
from sqlalchemy.orm import Mapped, mapped_column

from data_rentgen.db.models.base import Base


class Schema(Base):
    __tablename__ = "schema"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    digest: Mapped[UUID] = mapped_column(
        SQL_UUID,
        nullable=False,
        index=True,
        unique=True,
        doc="Schema SHA-1 digest based on fields content. Currently this is in form of UUID",
    )
    fields: Mapped[dict] = mapped_column(
        JSON,
        nullable=False,
        doc="Schema fields",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from enum import Enum

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset import (
    OpenLineageInputDataset,
    OpenLineageOutputDataset,
)
from data_rentgen.consumer.openlineage.job import OpenLineageJob
from data_rentgen.consumer.openlineage.run import OpenLineageRun


class OpenLineageRunEventType(str, Enum):
    """Supported values of run eventType.
    See [RunEvent](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    START = "START"
    RUNNING = "RUNNING"
    COMPLETE = "COMPLETE"
    ABORT = "ABORT"
    FAIL = "FAIL"
    OTHER = "OTHER"

    def __str__(self) -> str:
        return self.value


class OpenLineageRunEvent(OpenLineageBase):
    """RunEvent model.
    See [RunEvent](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    eventTime: datetime
    eventType: OpenLineageRunEventType
    job: OpenLineageJob
    run: OpenLineageRun
    inputs: list[OpenLineageInputDataset] = Field(default_factory=list)
    outputs: list[OpenLineageOutputDataset] = Field(default_factory=list)
    # ignore producer and schemaURL



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Literal

from data_rentgen.consumer.settings.security.base import KafkaSecurityBaseSettings


class KafkaSecurityAnonymousSettings(KafkaSecurityBaseSettings):
    """Kafka anonymous auth settings.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__KAFKA__SECURITY__TYPE=None
    """

    type: Literal[None] = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum, IntEnum
from uuid import UUID

from data_rentgen.dto.run import RunDTO


class OperationTypeDTO(str, Enum):
    BATCH = "BATCH"
    STREAMING = "STREAMING"

    def __str__(self) -> str:
        return str(self.value)


class OperationStatusDTO(IntEnum):
    UNKNOWN = -1
    STARTED = 0
    SUCCEEDED = 1
    FAILED = 2
    KILLED = 3


@dataclass
class OperationDTO:
    id: UUID
    run: RunDTO
    name: str
    type: OperationTypeDTO | None = None
    position: int | None = None
    group: str | None = None
    description: str | None = None
    status: OperationStatusDTO = OperationStatusDTO.UNKNOWN
    started_at: datetime | None = None
    ended_at: datetime | None = None

    @property
    def unique_key(self) -> tuple:
        return (self.id,)

    def merge(self, new: OperationDTO) -> OperationDTO:
        return OperationDTO(
            id=self.id,
            run=self.run.merge(new.run),
            name=new.name or self.name,
            type=new.type or self.type,
            group=new.group or self.group,
            description=new.description or self.description,
            status=max(new.status, self.status),
            position=new.position or self.position,
            started_at=new.started_at or self.started_at,
            ended_at=new.ended_at or self.ended_at,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create dataset

Revision ID: 2cb695c0a318
Revises: 81153f43e276
Create Date: 2024-06-27 19:10:39.585904

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "2cb695c0a318"
down_revision = "81153f43e276"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "dataset",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("location_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("format", sa.String(length=32), nullable=True),
        sa.Column(
            "search_vector",
            postgresql.TSVECTOR(),
            sa.Computed(
                "to_tsvector('simple'::regconfig, name || ' ' || (translate(name, '/.', '  ')))",
                persisted=True,
            ),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["location_id"],
            ["location.id"],
            name=op.f("fk__dataset__location_id__location"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__dataset")),
        sa.UniqueConstraint("location_id", "name", name=op.f("uq__dataset__location_id_name")),
    )
    op.create_index(op.f("ix__dataset__name"), "dataset", ["name"], unique=False)
    op.create_index(op.f("ix__dataset__location_id"), "dataset", ["location_id"], unique=False)
    op.create_index(
        op.f("ix__dataset__search_vector"),
        "dataset",
        ["search_vector"],
        unique=False,
        postgresql_using="gin",
    )


def downgrade() -> None:
    op.drop_index(op.f("ix__dataset__location_id"), table_name="dataset")
    op.drop_index(op.f("ix__dataset__name"), table_name="dataset")
    op.drop_index(op.f("ix__dataset__search_vector"), table_name="dataset", postgresql_using="gin")
    op.drop_table("dataset")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import http
from typing import Any

from pydantic import BaseModel
from typing_extensions import Literal

from data_rentgen.exceptions.entity import EntityNotFoundError
from data_rentgen.server.errors.base import BaseErrorSchema
from data_rentgen.server.errors.registration import register_error_response


class NotFoundDetailsSchema(BaseModel):
    entity_type: str
    field: str
    value: Any


@register_error_response(
    exception=EntityNotFoundError,
    status=http.HTTPStatus.NOT_FOUND,
)
class NotFoundSchema(BaseErrorSchema):
    code: Literal["not_found"] = "not_found"
    details: NotFoundDetailsSchema



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import http
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, ValidationError
from typing_extensions import Literal

from data_rentgen.server.errors.base import BaseErrorSchema
from data_rentgen.server.errors.registration import register_error_response


class InvalidRequestBaseErrorSchema(BaseModel):
    loc: list[str | int] = Field(alias="location")
    msg: str = Field(alias="message")
    type: str = Field(alias="code")
    ctx: dict = Field(default_factory=dict, alias="context")
    input: Any = Field(default=None)

    model_config = ConfigDict(populate_by_name=True)


@register_error_response(
    exception=ValidationError,
    status=http.HTTPStatus.UNPROCESSABLE_ENTITY,
)
class InvalidRequestSchema(BaseErrorSchema):
    code: Literal["invalid_request"] = "invalid_request"
    message: Literal["Invalid request"] = "Invalid request"
    details: list[InvalidRequestBaseErrorSchema]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets import (
    OpenLineageDatasetFacets,
    OpenLineageInputDatasetFacets,
    OpenLineageOutputDatasetFacets,
)


class OpenLineageDataset(OpenLineageBase):
    """Generic dataset model.
    See [Dataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    namespace: str = Field(json_schema_extra={"format": "uri"})
    name: str
    facets: OpenLineageDatasetFacets = Field(default_factory=OpenLineageDatasetFacets)


class OpenLineageInputDataset(OpenLineageDataset):
    """Input dataset model.
    See [InputDataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    inputFacets: OpenLineageInputDatasetFacets = Field(default_factory=OpenLineageInputDatasetFacets)


class OpenLineageOutputDataset(OpenLineageDataset):
    """Output dataset model.
    See [OutputDataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    outputFacets: OpenLineageOutputDatasetFacets = Field(default_factory=OpenLineageOutputDatasetFacets)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from enum import Enum

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet


class OpenLineageAirflowDagInfo(OpenLineageBase):
    """Airflow dag info.
    See [Dag](https://github.com/apache/airflow/blob/providers-openlineage/1.9.0/airflow/providers/openlineage/facets/AirflowRunFacet.json).
    """

    dag_id: str
    owner: str | None = None


class OpenLineageAirflowDagRunType(Enum):
    """Airflow dagRun type.
    See [DagRunType](https://github.com/apache/airflow/blob/2.9.3/airflow/utils/types.py#L48-L54).
    """

    BACKFILL_JOB = "backfill"
    SCHEDULED = "scheduled"
    MANUAL = "manual"
    DATASET_TRIGGERED = "dataset_triggered"


class OpenLineageAirflowDagRunInfo(OpenLineageBase):
    """Airflow dagRun info.
    See [DagRun](https://github.com/apache/airflow/blob/providers-openlineage/1.9.0/airflow/providers/openlineage/facets/AirflowRunFacet.json).
    """

    run_id: str
    run_type: OpenLineageAirflowDagRunType
    data_interval_start: datetime
    data_interval_end: datetime


class OpenLineageAirflowTaskInfo(OpenLineageBase):
    """Airflow task info.
    See [Task](https://github.com/apache/airflow/blob/providers-openlineage/1.9.0/airflow/providers/openlineage/facets/AirflowRunFacet.json).
    """

    task_id: str


class OpenLineageAirflowTaskInstanceInfo(OpenLineageBase):
    """Airflow taskInstance info.
    See [TaskInstance](https://github.com/apache/airflow/blob/providers-openlineage/1.9.0/airflow/providers/openlineage/facets/AirflowRunFacet.json).
    """

    try_number: int
    map_index: int | None = None
    log_url: str | None = None


class OpenLineageAirflowDagRunFacet(OpenLineageRunFacet):
    """Run facet describing Airflow DAG run.
    See [AirflowDagRunFacet](https://github.com/apache/airflow/blob/main/airflow/providers/openlineage/facets/AirflowDagRunFacet.json).
    """

    dag: OpenLineageAirflowDagInfo
    dagRun: OpenLineageAirflowDagRunInfo


class OpenLineageAirflowTaskRunFacet(OpenLineageRunFacet):
    """Run facet describing Airflow Task run.
    See [AirflowRunFacet](https://github.com/apache/airflow/blob/providers-openlineage/1.9.0/airflow/providers/openlineage/facets/AirflowRunFacet.json).
    """

    dag: OpenLineageAirflowDagInfo
    dagRun: OpenLineageAirflowDagRunInfo
    task: OpenLineageAirflowTaskInfo
    taskInstance: OpenLineageAirflowTaskInstanceInfo



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.base import OpenLineageBase


class OpenLineageRunFacet(OpenLineageBase):
    """Base class for all run facets.
    See [RunFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.extractors.run import extract_parent_run
from data_rentgen.consumer.openlineage.run_event import (
    OpenLineageRunEvent,
    OpenLineageRunEventType,
)
from data_rentgen.dto import OperationDTO, OperationStatusDTO, OperationTypeDTO


def extract_operation(event: OpenLineageRunEvent) -> OperationDTO:
    # operation always has parent
    run = extract_parent_run(event.run.facets.parent)  # type: ignore[arg-type]

    # in some cases, operation name may contain raw SELECT query with newlines
    operation_name = " ".join(line.strip() for line in event.job.name.splitlines()).strip()
    # remove parent job name from operation name
    if operation_name.startswith(run.job.name):
        prefix = len(run.job.name) + 1
        operation_name = operation_name[prefix:]

    operation = OperationDTO(
        id=event.run.runId,  # type: ignore [arg-type]
        run=run,
        name=operation_name,
        type=OperationTypeDTO(event.job.facets.jobType.processingType) if event.job.facets.jobType else None,
    )
    enrich_operation_status(operation, event)
    enrich_operation_description(operation, event)
    return operation


def enrich_operation_status(operation: OperationDTO, event: OpenLineageRunEvent) -> OperationDTO:
    match event.eventType:
        case OpenLineageRunEventType.START:
            operation.started_at = event.eventTime
            operation.status = OperationStatusDTO.STARTED
        case OpenLineageRunEventType.RUNNING:
            operation.status = OperationStatusDTO.STARTED
        case OpenLineageRunEventType.COMPLETE:
            operation.ended_at = event.eventTime
            operation.status = OperationStatusDTO.SUCCEEDED
        case OpenLineageRunEventType.FAIL:
            operation.ended_at = event.eventTime
            operation.status = OperationStatusDTO.FAILED
        case OpenLineageRunEventType.ABORT:
            operation.ended_at = event.eventTime
            operation.status = OperationStatusDTO.KILLED
        case OpenLineageRunEventType.OTHER:
            # OTHER is used only to update run statistics
            pass
    return operation


def enrich_operation_description(operation: OperationDTO, event: OpenLineageRunEvent) -> OperationDTO:
    spark_job_details = event.run.facets.spark_jobDetails
    if spark_job_details:
        operation.position = spark_job_details.jobId
        operation.group = spark_job_details.jobGroup
        operation.description = spark_job_details.jobDescription
    return operation



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.db.models.address import Address
from data_rentgen.db.models.base import Base
from data_rentgen.db.models.column_lineage import ColumnLineage
from data_rentgen.db.models.custom_properties import CustomProperties
from data_rentgen.db.models.custom_user_properties import CustomUserProperties
from data_rentgen.db.models.dataset import Dataset
from data_rentgen.db.models.dataset_column_relation import (
    DatasetColumnRelation,
    DatasetColumnRelationType,
)
from data_rentgen.db.models.dataset_symlink import DatasetSymlink, DatasetSymlinkType
from data_rentgen.db.models.input import Input
from data_rentgen.db.models.job import Job, JobType
from data_rentgen.db.models.location import Location
from data_rentgen.db.models.operation import Operation, OperationStatus, OperationType
from data_rentgen.db.models.output import Output, OutputType
from data_rentgen.db.models.run import Run, RunStartReason, RunStatus
from data_rentgen.db.models.schema import Schema
from data_rentgen.db.models.user import User

__all__ = [
    "Address",
    "Base",
    "ColumnLineage",
    "CustomProperties",
    "CustomUserProperties",
    "Dataset",
    "DatasetColumnRelation",
    "DatasetColumnRelationType",
    "DatasetSymlink",
    "DatasetSymlinkType",
    "Input",
    "Job",
    "JobType",
    "Location",
    "Operation",
    "OperationStatus",
    "OperationType",
    "Output",
    "OutputType",
    "Run",
    "RunStartReason",
    "RunStatus",
    "Schema",
    "User",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class UserDTO:
    name: str
    id: int | None = field(default=None, compare=False)

    @property
    def unique_key(self) -> tuple:
        return (self.name,)

    def merge(self, new: UserDTO) -> UserDTO:
        return UserDTO(
            name=self.name,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Annotated, Any, NoReturn

from fastapi import Depends, FastAPI, Request
from jwcrypto.common import JWException
from keycloak import KeycloakOpenID, KeycloakOperationError

from data_rentgen.db.models import User
from data_rentgen.dependencies import Stub
from data_rentgen.dto import UserDTO
from data_rentgen.exceptions.auth import AuthorizationError, LogoutError
from data_rentgen.exceptions.redirect import RedirectError
from data_rentgen.server.providers.auth.base_provider import AuthProvider
from data_rentgen.server.settings.auth.keycloak import KeycloakAuthProviderSettings
from data_rentgen.services import UnitOfWork

logger = logging.getLogger(__name__)


class KeycloakAuthProvider(AuthProvider):
    def __init__(
        self,
        settings: Annotated[KeycloakAuthProviderSettings, Depends(Stub(KeycloakAuthProviderSettings))],
        unit_of_work: Annotated[UnitOfWork, Depends()],
    ) -> None:
        self.settings = settings
        self._uow = unit_of_work
        self.keycloak_openid = KeycloakOpenID(
            server_url=self.settings.keycloak.server_url,
            client_id=self.settings.keycloak.client_id,
            realm_name=self.settings.keycloak.realm_name,
            client_secret_key=self.settings.keycloak.client_secret.get_secret_value(),
            verify=self.settings.keycloak.verify_ssl,
        )

    @classmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        settings = KeycloakAuthProviderSettings.model_validate(
            app.state.settings.auth.model_dump(exclude={"provider"}),
        )
        logger.info("Using %s provider with settings:\n%s", cls.__name__, settings)
        app.dependency_overrides[AuthProvider] = cls
        app.dependency_overrides[KeycloakAuthProviderSettings] = lambda: settings
        return app

    async def get_token_password_grant(
        self,
        login: str,
        password: str,
    ) -> dict[str, Any]:
        msg = "Password grant is not supported by KeycloakAuthProvider."
        raise NotImplementedError(msg)

    async def get_token_authorization_code_grant(
        self,
        code: str,
    ) -> dict[str, Any]:
        try:
            return await self.keycloak_openid.a_token(
                grant_type="authorization_code",
                code=code,
                redirect_uri=self.settings.keycloak.redirect_uri,
            )
        except KeycloakOperationError as e:
            logger.exception("Fail to get token from keycloak: %s", e)  # noqa: TRY401
            msg = "Failed to get token"
            raise AuthorizationError(msg) from e

    async def get_current_user(
        self,
        access_token: str | None,
        request: Request,
    ) -> User:
        if not access_token:
            logger.debug("No access token found in session.")
            await self.redirect_to_auth()

        # if user is disabled or blocked in Keycloak after the token is issued, he will
        # remain authorized until the token expires (not more than 15 minutes in MTS SSO)
        token_info = await self.decode_token(access_token)
        refresh_token = request.session.get("refresh_token")

        if token_info is None and refresh_token:
            logger.debug("Access token invalid. Attempting to refresh.")
            access_token, refresh_token = await self.refresh_access_token(refresh_token)
            token_info = await self.decode_token(access_token)
            request.session["access_token"] = access_token
            request.session["refresh_token"] = refresh_token

        if token_info is None:
            await self.redirect_to_auth()

        # this name is hardcoded in keycloak:
        # https://github.com/keycloak/keycloak/blob/3ca3a4ad349b4d457f6829eaf2ae05f1e01408be/core/src/main/java/org/keycloak/representations/IDToken.java
        login = token_info.get("preferred_username")  # type: ignore[union-attr]
        if not login:
            msg = "Invalid token payload"
            raise AuthorizationError(msg)
        return await self._uow.user.get_or_create(UserDTO(name=login))  # type: ignore[arg-type]

    async def decode_token(self, access_token: str) -> dict[str, Any] | None:
        try:
            return await self.keycloak_openid.a_decode_token(token=access_token)
        except (KeycloakOperationError, JWException) as err:
            logger.debug("Access token is invalid or expired: %s", err)
            return None

    async def refresh_access_token(self, refresh_token: str) -> tuple[str, str]:  # type: ignore[return]
        try:
            new_tokens = await self.keycloak_openid.a_refresh_token(refresh_token)
            logger.debug("Access token refreshed")
            return new_tokens["access_token"], new_tokens["refresh_token"]
        except (KeycloakOperationError, JWException) as err:
            logger.debug("Failed to refresh access token: %s", err)
            await self.redirect_to_auth()

    async def redirect_to_auth(self) -> NoReturn:
        auth_url = await self.keycloak_openid.a_auth_url(
            redirect_uri=self.settings.keycloak.redirect_uri,
            scope=self.settings.keycloak.scope,
        )

        logger.debug("Redirecting user to auth url: %s", auth_url)
        raise RedirectError(
            message="Please authorize using provided URL",
            details=auth_url,
        )

    async def logout(self, user: User, refresh_token: str | None) -> None:
        if not refresh_token:
            logger.debug("No refresh token found in session.")
            return

        try:
            await self.keycloak_openid.a_logout(refresh_token)
        except KeycloakOperationError as err:
            msg = f"Can't logout user: {user.name}"
            logger.debug("%s. Error: %s", msg, err)
            raise LogoutError(msg) from err



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

DEFAULT_MAX_AGE = 1_209_600


class SessionSettings(BaseModel):
    """Session Middleware Settings.

    See `SessionMiddleware <https://www.starlette.io/middleware/#sessionmiddleware>`_ documentation.

    .. note::

        You can pass here any extra option supported by ``SessionMiddleware``,
        even if it is not mentioned in documentation.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__SERVER__SESSION__SECRET_KEY=secret
        DATA_RENTGEN__SERVER__SESSION__SESSION_COOKIE=custom_cookie_name
        DATA_RENTGEN__SERVER__SESSION__MAX_AGE=None  # cookie will last as long as the browser session
        DATA_RENTGEN__SERVER__SESSION__SAME_SITE=strict
        DATA_RENTGEN__SERVER__SESSION__HTTPS_ONLY=True
        DATA_RENTGEN__SERVER__SESSION__DOMAIN=example.com

    """

    secret_key: str = Field(description="A random string for signing cookies.")
    session_cookie: str | None = Field(default="session", description="Name of the session cookie.")
    max_age: int | None = Field(
        default=DEFAULT_MAX_AGE,
        description="Session expiry time in seconds. Defaults to 2 weeks.",
    )
    same_site: str | None = Field(
        default="lax",
        description="Prevents cookie from being sent with cross-site requests.",
    )
    path: str | None = Field(default="/", description="Path to restrict session cookie access.")
    https_only: bool = Field(default=False, description="Secure flag for HTTPS-only access.")
    domain: str | None = Field(
        default=None,
        description="Domain for sharing cookies between subdomains or cross-domains.",
    )

    model_config = ConfigDict(extra="allow")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Sequence
from enum import IntFlag
from string import punctuation

from sqlalchemy import ColumnElement, func
from sqlalchemy.orm import InstrumentedAttribute

# left some punctuation to match file paths, URLs and host names
TSQUERY_UNSUPPORTED_CHARS = "".join(sorted(set(punctuation) - {"/", ".", "_", "-"}))
TSQUERY_UNSUPPORTED_CHARS_REPLACEMENT = str.maketrans(TSQUERY_UNSUPPORTED_CHARS, " " * len(TSQUERY_UNSUPPORTED_CHARS))
TSQUERY_ALL_PUNCTUATION_REPLACEMENT = str.maketrans(punctuation, " " * len(punctuation))


class SearchRankNormalization(IntFlag):
    """See https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-RANKING"""

    IGNORE_LENGTH = 0
    DOCUMENT_LENGTH_LOGARITHM = 1
    DOCUMENT_LENGTH = 2
    HARMONIC_DISTANCE = 4
    UNIQUE_WORDS = 8
    UNIQUE_WORDS_LOGARITHM = 16
    RANK_PLUS_ONE = 32


def ts_rank(search_vector: InstrumentedAttribute, ts_query: ColumnElement) -> ColumnElement:
    """Get ts_rank for search query ranking.

    Places results with smaller number of total words (like table name) to the top,
    and long results (as file paths) to the bottom.

    Also places on top results with lexemes order matching the tsvector order.
    """
    return func.ts_rank_cd(search_vector, ts_query, SearchRankNormalization.UNIQUE_WORDS)


def make_tsquery(user_input: str) -> ColumnElement:
    """Convert user input to tsquery.

    Work almost like `plainto_tsquery`, but wraps words with `:*` to allow matching by prefix,
    and also allows to match file paths or domain names by parts separated with dot.
    """
    # 'english' dictionary performs stemming, converting `application_123_123` to tsvector `'applic':1 '123':2`,
    # which does not match a tsquery like 'application:* & 123:*'.
    # Instead prefer 'simple' dictionary as it does not use stemming.
    return func.to_tsquery("simple", build_tsquery(user_input))


def ts_match(search_vector: InstrumentedAttribute, ts_query: ColumnElement) -> ColumnElement:
    """Build an expression to get only search_vector matching ts_query."""
    return search_vector.op("@@")(ts_query)


def build_tsquery(user_input: str) -> str:
    original_words = words_with_supported_punctuation(user_input)
    only_words = words_without_any_punctuation(user_input)

    return combine_queries(
        combine_words(*original_words, by_prefix=False),
        combine_words(*original_words),
        combine_words(*only_words) if only_words != original_words else [],
    )


def combine_words(*words: Sequence[str], by_prefix: bool = True) -> str:
    # Convert this ['some', 'query']
    # to this `'some' & 'query'` or `'some':* & 'query':*`
    modifier = ":*" if by_prefix else ""
    return " & ".join(f"'{word}'{modifier}" for word in words if word)


def combine_queries(*queries: Sequence[str]) -> str:
    # Convert this ['/some/file/path:* & abc:*', 'some:* & file:* & path:*']
    # to this '(/some/file/path:* & abc:*) | (some:* & file:* & path:* & abc:*)'
    return " | ".join(f"({query})" for query in queries if query)


def words_with_supported_punctuation(query: str) -> list[str]:
    # convert '@/some/path.or.domain!' -> '/some/path.or.domain'
    converted = query.translate(TSQUERY_UNSUPPORTED_CHARS_REPLACEMENT)
    result = []
    for part in converted.split():
        # delete parts containing only punctuation chars, like ./
        if not all(char in punctuation for char in part):
            result.append(part)  # noqa: PERF401
    return result


def words_without_any_punctuation(query: str) -> list[str]:
    # convert '@/some/path.or.domain!' -> 'some path or domain'
    return query.translate(TSQUERY_ALL_PUNCTUATION_REPLACEMENT).split()



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageDocumentationDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing documentation.
    See [DocumentationDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/DocumentationDatasetFacet.json).
    """

    description: str



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from enum import Enum

from pydantic import field_validator

from data_rentgen.consumer.openlineage.job_facets.base import OpenLineageJobFacet


class OpenLineageJobIntegrationType(str, Enum):
    """Integration where job is running.
    See [JobTypeJobFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/JobTypeJobFacet.json).
    """

    SPARK = "SPARK"
    AIRFLOW = "AIRFLOW"

    def __str__(self) -> str:
        return self.value


class OpenLineageJobType(str, Enum):
    """Job type.
    See [JobTypeJobFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/JobTypeJobFacet.json).
    """

    APPLICATION = "APPLICATION"
    JOB = "JOB"
    DAG = "DAG"
    TASK = "TASK"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, value):
        if value in {"SQL_JOB", "RDD_JOB"}:
            return cls.JOB
        return None


class OpenLineageJobProcessingType(str, Enum):
    """Job processing type.
    See [JobTypeJobFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/JobTypeJobFacet.json).
    """

    BATCH = "BATCH"
    STREAMING = "STREAMING"


class OpenLineageJobTypeJobFacet(OpenLineageJobFacet):
    """Job facet describing job type.
    See [JobTypeJobFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/JobTypeJobFacet.json).
    """

    integration: OpenLineageJobIntegrationType
    jobType: OpenLineageJobType
    processingType: OpenLineageJobProcessingType | None = None

    @field_validator("processingType", mode="before")
    @classmethod
    def _validate_processing_type(cls, processing_type: str):
        if processing_type == "NONE":
            return None
        return processing_type



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import AsyncGenerator

from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_engine_from_config,
    async_sessionmaker,
)

from data_rentgen.db.settings import DatabaseSettings


def create_session_factory(settings: DatabaseSettings) -> async_sessionmaker[AsyncSession]:
    engine = async_engine_from_config(settings.model_dump(), prefix="")
    return async_sessionmaker(
        bind=engine,
        class_=AsyncSession,
        expire_on_commit=False,
    )


def session_generator(settings: DatabaseSettings):
    a_session = create_session_factory(settings)

    async def wrapper() -> AsyncGenerator[AsyncSession, None]:
        async with a_session.begin() as session:
            yield session

    return wrapper



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from sqlalchemy import BigInteger, String
from sqlalchemy.orm import Mapped, mapped_column

from data_rentgen.db.models.base import Base


class User(Base):
    __tablename__ = "user"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    name: Mapped[str] = mapped_column(String, index=True, unique=True)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Collection
from datetime import datetime
from typing import NamedTuple
from uuid import UUID

from sqlalchemy import ColumnElement, any_, func, select
from sqlalchemy.dialects.postgresql import insert

from data_rentgen.db.models import ColumnLineage, DatasetColumnRelation
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.uuid import (
    extract_timestamp_from_uuid,
    generate_incremental_uuid,
)
from data_rentgen.dto import ColumnLineageDTO


class ColumnLineageRow(NamedTuple):
    source_dataset_id: int
    target_dataset_id: int
    source_column: str
    target_column: str
    types_combined: int
    last_used_at: datetime


class ColumnLineageRepository(Repository[ColumnLineage]):
    def get_id(self, item: ColumnLineageDTO) -> UUID:
        # `created_at' field of lineage should be the same as operation's,
        # to avoid scanning all partitions and speed up queries
        created_at = extract_timestamp_from_uuid(item.operation.id)

        # instead of using UniqueConstraint on multiple fields, which may produce a lot of table scans,
        # use them to calculate unique id
        id_components = [
            str(item.operation.id),
            str(item.source_dataset.id),
            str(item.target_dataset.id),
            str(item.fingerprint),
        ]
        return generate_incremental_uuid(created_at, ".".join(id_components))

    async def create_bulk(self, items: list[ColumnLineageDTO]):
        if not items:
            return

        insert_statement = insert(ColumnLineage)
        statement = insert_statement.on_conflict_do_nothing(
            index_elements=[ColumnLineage.created_at, ColumnLineage.id],
        )

        await self._session.execute(
            statement,
            [
                {
                    "id": self.get_id(item),
                    "created_at": extract_timestamp_from_uuid(item.operation.id),
                    "operation_id": item.operation.id,
                    "run_id": item.operation.run.id,
                    "job_id": item.operation.run.job.id,  # type: ignore[arg-type]
                    "source_dataset_id": item.source_dataset.id,
                    "target_dataset_id": item.target_dataset.id,
                    "fingerprint": item.fingerprint,
                }
                for item in items
            ],
        )

    async def list_by_job_ids(
        self,
        job_ids: Collection[int],
        since: datetime,
        until: datetime | None,
        source_dataset_ids: Collection[int],
        target_dataset_ids: Collection[int],
    ):
        if not job_ids:
            return []

        where = [
            ColumnLineage.created_at >= since,
            ColumnLineage.job_id == any_(list(job_ids)),  # type: ignore[arg-type]
            ColumnLineage.source_dataset_id == any_(list(source_dataset_ids)),  # type: ignore[arg-type]
            ColumnLineage.target_dataset_id == any_(list(target_dataset_ids)),  # type: ignore[arg-type]
        ]
        if until:
            where.append(ColumnLineage.created_at <= until)

        return await self._get_column_lineage_with_column_relations(where)

    async def list_by_run_ids(
        self,
        run_ids: Collection[UUID],
        since: datetime,
        until: datetime | None,
        source_dataset_ids: Collection[int],
        target_dataset_ids: Collection[int],
    ):
        if not run_ids:
            return []

        where = [
            ColumnLineage.created_at >= since,
            ColumnLineage.run_id == any_(list(run_ids)),  # type: ignore[arg-type]
            ColumnLineage.source_dataset_id == any_(list(source_dataset_ids)),  # type: ignore[arg-type]
            ColumnLineage.target_dataset_id == any_(list(target_dataset_ids)),  # type: ignore[arg-type]
        ]
        if until:
            where.append(ColumnLineage.created_at <= until)
        return await self._get_column_lineage_with_column_relations(where)

    async def list_by_operation_ids(
        self,
        operation_ids: Collection[UUID],
        source_dataset_ids: Collection[int],
        target_dataset_ids: Collection[int],
    ):
        if not operation_ids:
            return []

        # Output created_at is always the same as operation's created_at
        # do not use `tuple_(Output.created_at, Output.operation_id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        min_created_at = extract_timestamp_from_uuid(min(operation_ids))
        max_created_at = extract_timestamp_from_uuid(max(operation_ids))
        where = [
            ColumnLineage.created_at >= min_created_at,
            ColumnLineage.created_at <= max_created_at,
            ColumnLineage.operation_id == any_(list(operation_ids)),  # type: ignore[arg-type]
            ColumnLineage.source_dataset_id == any_(list(source_dataset_ids)),  # type: ignore[arg-type]
            ColumnLineage.target_dataset_id == any_(list(target_dataset_ids)),  # type: ignore[arg-type]
        ]
        return await self._get_column_lineage_with_column_relations(where)

    async def _get_column_lineage_with_column_relations(self, where: Collection[ColumnElement]):
        query = (
            select(
                ColumnLineage.source_dataset_id,
                ColumnLineage.target_dataset_id,
                DatasetColumnRelation.source_column,
                DatasetColumnRelation.target_column,
                func.bit_or(DatasetColumnRelation.type).label("types_combined"),
                func.max(ColumnLineage.created_at).label("last_used_at"),
            )
            .join(DatasetColumnRelation, ColumnLineage.fingerprint == DatasetColumnRelation.fingerprint)
            .group_by(
                ColumnLineage.source_dataset_id,
                ColumnLineage.target_dataset_id,
                DatasetColumnRelation.source_column,
                DatasetColumnRelation.target_column,
            )
        )

        query = query.where(*where)
        result = await self._session.execute(query)
        return [
            ColumnLineageRow(
                source_dataset_id=row.source_dataset_id,
                target_dataset_id=row.target_dataset_id,
                source_column=row.source_column,
                target_column=row.target_column,
                types_combined=row.types_combined,
                last_used_at=row.last_used_at,
            )
            for row in result.all()
        ]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.run_facets.airflow import (
    OpenLineageAirflowDagInfo,
    OpenLineageAirflowDagRunFacet,
    OpenLineageAirflowDagRunInfo,
    OpenLineageAirflowDagRunType,
    OpenLineageAirflowTaskInfo,
    OpenLineageAirflowTaskInstanceInfo,
    OpenLineageAirflowTaskRunFacet,
)
from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet
from data_rentgen.consumer.openlineage.run_facets.parent_run import (
    OpenLineageParentJob,
    OpenLineageParentRun,
    OpenLineageParentRunFacet,
)
from data_rentgen.consumer.openlineage.run_facets.processing_engine import (
    OpenLineageProcessingEngineName,
    OpenLineageProcessingEngineRunFacet,
)
from data_rentgen.consumer.openlineage.run_facets.spark_application import (
    OpenLineageSparkApplicationDetailsRunFacet,
    OpenLineageSparkDeployMode,
)
from data_rentgen.consumer.openlineage.run_facets.spark_job import (
    OpenLineageSparkJobDetailsRunFacet,
)

__all__ = [
    "OpenLineageAirflowDagInfo",
    "OpenLineageAirflowDagRunFacet",
    "OpenLineageAirflowDagRunInfo",
    "OpenLineageAirflowDagRunType",
    "OpenLineageAirflowTaskInfo",
    "OpenLineageAirflowTaskInstanceInfo",
    "OpenLineageAirflowTaskRunFacet",
    "OpenLineageParentJob",
    "OpenLineageParentRun",
    "OpenLineageParentRunFacet",
    "OpenLineageProcessingEngineName",
    "OpenLineageProcessingEngineRunFacet",
    "OpenLineageRunFacet",
    "OpenLineageRunFacets",
    "OpenLineageSparkApplicationDetailsRunFacet",
    "OpenLineageSparkDeployMode",
    "OpenLineageSparkJobDetailsRunFacet",
]


class OpenLineageRunFacets(OpenLineageBase):
    """All possible run facets.
    See [Run](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    parent: OpenLineageParentRunFacet | None = None
    processing_engine: OpenLineageProcessingEngineRunFacet | None = None
    spark_applicationDetails: OpenLineageSparkApplicationDetailsRunFacet | None = None
    spark_jobDetails: OpenLineageSparkJobDetailsRunFacet | None = None
    airflow: OpenLineageAirflowTaskRunFacet | None = None
    airflowDagRun: OpenLineageAirflowDagRunFacet | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create dataset_symlink

Revision ID: c2c583123133
Revises: 2cb695c0a318
Create Date: 2024-06-27 19:10:11.686319

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c2c583123133"
down_revision = "2cb695c0a318"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "dataset_symlink",
        sa.Column("id", sa.BigInteger(), autoincrement=True, nullable=False),
        sa.Column("from_dataset_id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("to_dataset_id", sa.BigInteger(), autoincrement=False, nullable=False),
        sa.Column("type", sa.String(length=32), nullable=False),
        sa.ForeignKeyConstraint(
            ["from_dataset_id"],
            ["dataset.id"],
            name=op.f("fk__dataset_symlink__from_dataset_id__dataset"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["to_dataset_id"],
            ["dataset.id"],
            name=op.f("fk__dataset_symlink__to_dataset_id__dataset"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__dataset_symlink")),
        sa.UniqueConstraint(
            "from_dataset_id",
            "to_dataset_id",
            name=op.f("uq__dataset_symlink__from_dataset_id_to_dataset_id"),
        ),
    )
    op.create_index(op.f("ix__dataset_symlink__to_dataset_id"), "dataset_symlink", ["to_dataset_id"], unique=False)
    op.create_index(op.f("ix__dataset_symlink__from_dataset_id"), "dataset_symlink", ["from_dataset_id"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix__dataset_symlink__from_dataset_id"), table_name="dataset_symlink")
    op.drop_index(op.f("ix__dataset_symlink__to_dataset_id"), table_name="dataset_symlink")
    op.drop_table("dataset_symlink")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from dataclasses import dataclass
from typing import Annotated

from fastapi import Depends
from sqlalchemy import Row

from data_rentgen.db.models.location import Location
from data_rentgen.dto.pagination import PaginationDTO
from data_rentgen.services.uow import UnitOfWork


@dataclass
class LocationServiceDatasetStatistics:
    total_datasets: int = 0

    @classmethod
    def from_row(cls, row: Row | None):
        if not row:
            return cls()

        return cls(total_datasets=row.total_datasets)


@dataclass
class LocationServiceJobStatistics:
    total_jobs: int = 0

    @classmethod
    def from_row(cls, row: Row | None):
        if not row:
            return cls()

        return cls(total_jobs=row.total_jobs)


@dataclass
class LocationServiceStatistics:
    datasets: LocationServiceDatasetStatistics
    jobs: LocationServiceJobStatistics


@dataclass
class LocationServiceResult:
    id: int
    data: Location
    statistics: LocationServiceStatistics


class LocationServicePaginatedResult(PaginationDTO[LocationServiceResult]):
    pass


class LocationService:
    def __init__(self, uow: Annotated[UnitOfWork, Depends()]):
        self._uow = uow

    async def paginate(
        self,
        page: int,
        page_size: int,
        location_ids: Collection[int],
        location_type: str | None,
        search_query: str | None,
    ) -> LocationServicePaginatedResult:
        pagination = await self._uow.location.paginate(
            page=page,
            page_size=page_size,
            location_ids=location_ids,
            location_type=location_type,
            search_query=search_query,
        )
        location_ids = [item.id for item in pagination.items]
        dataset_stats = await self._uow.dataset.get_stats_by_location_ids(location_ids)
        job_stats = await self._uow.job.get_stats_by_location_ids(location_ids)

        return LocationServicePaginatedResult(
            page=pagination.page,
            page_size=pagination.page_size,
            total_count=pagination.total_count,
            items=[
                LocationServiceResult(
                    id=location.id,
                    data=location,
                    statistics=LocationServiceStatistics(
                        datasets=LocationServiceDatasetStatistics.from_row(dataset_stats.get(location.id)),
                        jobs=LocationServiceJobStatistics.from_row(job_stats.get(location.id)),
                    ),
                )
                for location in pagination.items
            ],
        )

    async def update_external_id(
        self,
        location_id: int,
        external_id: str | None,
    ) -> LocationServiceResult:
        location = await self._uow.location.update_external_id(location_id, external_id)
        dataset_stats = await self._uow.dataset.get_stats_by_location_ids([location.id])
        job_stats = await self._uow.job.get_stats_by_location_ids([location.id])
        return LocationServiceResult(
            id=location.id,
            data=location,
            statistics=LocationServiceStatistics(
                datasets=LocationServiceDatasetStatistics.from_row(dataset_stats.get(location.id)),
                jobs=LocationServiceJobStatistics.from_row(job_stats.get(location.id)),
            ),
        )
